# Copyright (C) 2014-2020 The BET Development Team

import unittest
import os
import glob
import numpy as np
import numpy.testing as nptest
import bet
import bet.sample as sample
import bet.util as util
import bet.sampling.basicSampling as bsam
from bet.Comm import comm, MPI

#local_path = os.path.join(os.path.dirname(bet.__file__), "/test")
local_path = ''


class Test_sample_set(unittest.TestCase):
    def setUp(self):
        self.dim = 2
        self.num = 100
        self.values = np.ones((self.num, self.dim))
        self.sam_set = sample.sample_set(dim=self.dim)
        self.sam_set.set_values(self.values)
        self.domain = np.array([[0, 1], [0, 1]], dtype=np.float)

    @unittest.skipIf(comm.size > 1, 'Only run in serial')
    def test_merge(self):
        """
        Test merge.
        """
        other_set = self.sam_set.copy()
        merge_set = self.sam_set.merge(other_set)
        nptest.assert_array_equal(self.sam_set._domain, merge_set._domain)
        nptest.assert_array_equal(self.sam_set._values,
                                  merge_set._values[0:self.num, :])

    def test_normalize(self):
        """
        Test normalize and undo normalize domain.
        """
        domain = 5.0 * self.domain - 1.0
        self.sam_set.set_domain(domain)
        ee = np.ones((self.num, self.dim))
        self.sam_set.set_error_estimates(ee)
        jac = np.ones((self.num, 3, self.dim))
        self.sam_set.set_jacobians(jac)

        self.sam_set.normalize_domain()
        nptest.assert_array_equal(self.sam_set._domain, self.domain)
        nptest.assert_array_almost_equal(self.sam_set._values, 0.4)
        nptest.assert_array_almost_equal(self.sam_set._error_estimates, 0.4)
        nptest.assert_array_almost_equal(self.sam_set._jacobians, 5.0)

        self.sam_set.undo_normalize_domain()
        nptest.assert_array_equal(self.sam_set._domain, domain)
        nptest.assert_array_almost_equal(self.sam_set._values, 1.0)
        nptest.assert_array_almost_equal(self.sam_set._error_estimates, 1.0)
        nptest.assert_array_almost_equal(self.sam_set._jacobians, 1.0)

    def test_clip(self):
        """
        Test clipping of sample set.
        """
        ee = np.ones((self.num, self.dim))
        self.sam_set.set_error_estimates(ee)
        jac = np.ones((self.num, 3, self.dim))
        self.sam_set.set_jacobians(jac)

        cnum = int(0.5 * self.num)
        sam_set_clipped = self.sam_set.clip(cnum)

        num = sam_set_clipped.check_num()
        self.assertEqual(num, cnum)
        nptest.assert_array_equal(self.sam_set._values[0:cnum, :],
                                  sam_set_clipped._values)
        nptest.assert_array_equal(self.sam_set._error_estimates[0:cnum, :],
                                  sam_set_clipped._error_estimates)
        nptest.assert_array_equal(self.sam_set._jacobians[0:cnum, :],
                                  sam_set_clipped._jacobians)

    def test_set_domain(self):
        """
        Test set domain.
        """
        self.sam_set.set_domain(self.domain)
        nptest.assert_array_equal(self.sam_set._domain, self.domain)

    def test_get_domain(self):
        """
        Test get domain.
        """
        self.sam_set.set_domain(self.domain)
        nptest.assert_array_equal(self.sam_set.get_domain(), self.domain)

    @unittest.skipIf(comm.size > 1, 'Only run in serial')
    def test_save_load(self):
        """
        Check save_sample_set and load_sample_set.
        """
        prob = 1.0 / float(self.num) * np.ones((self.num,))
        self.sam_set.set_probabilities(prob)
        vol = 1.0 / float(self.num) * np.ones((self.num,))
        self.sam_set.set_volumes(vol)
        ee = np.ones((self.num, self.dim))
        self.sam_set.set_error_estimates(ee)
        jac = np.ones((self.num, 3, self.dim))
        self.sam_set.set_jacobians(jac)
        self.sam_set.global_to_local()
        self.sam_set.set_domain(self.domain)
        self.sam_set.update_bounds()
        self.sam_set.update_bounds_local()

        file_name = os.path.join(local_path, 'testfile')
        globalize = True
        util.save_object(self.sam_set, file_name, globalize)
        comm.barrier()
        loaded_set = util.load_object(file_name)
        assert self.sam_set == loaded_set

        if comm.rank == 0 and globalize:
            os.remove(file_name+'.p')
        elif not globalize:
            os.remove(file_name+'.p')

        comm.barrier()

        file_name = os.path.join(local_path, 'testfile')
        globalize = False
        util.save_object(self.sam_set, file_name, globalize)
        comm.barrier()

        if comm.size > 1 and not globalize:
            local_file_name = os.path.os.path.join(os.path.dirname(file_name),
                                                   "proc{}_{}".format(comm.rank, os.path.basename(file_name)))

        loaded_set = util.load_object(file_name)

        assert loaded_set == self.sam_set

        # Cleanup
        if comm.size == 1 or globalize:
            if comm.rank == 0:
                os.remove(file_name+'.p')
        else:
            os.remove(local_file_name+'.p')

    def test_copy(self):
        """
        Check copy.
        """
        prob = 1.0 / float(self.num) * np.ones((self.num,))
        self.sam_set.set_probabilities(prob)
        vol = 1.0 / float(self.num) * np.ones((self.num,))
        self.sam_set.set_volumes(vol)
        ee = np.ones((self.num, self.dim))
        self.sam_set.set_error_estimates(ee)
        jac = np.ones((self.num, 3, self.dim))
        self.sam_set.set_jacobians(jac)
        self.sam_set.global_to_local()
        self.sam_set.set_domain(self.domain)
        self.sam_set.update_bounds()
        self.sam_set.update_bounds_local()
        self.sam_set.set_kdtree()

        copied_set = self.sam_set.copy()
        if comm.size == 0:
            assert copied_set == self.sam_set

    def test_update_bounds(self):
        """
        Check update_bounds
        """
        self.sam_set.set_domain(self.domain)
        self.sam_set.update_bounds()
        nptest.assert_array_equal(self.sam_set._left,
                                  np.repeat([self.domain[:, 0]], self.num, 0))
        nptest.assert_array_equal(self.sam_set._right,
                                  np.repeat([self.domain[:, 1]], self.num, 0))
        nptest.assert_array_equal(self.sam_set._width,
                                  np.repeat([self.domain[:, 1] - self.domain[:, 0]], self.num, 0))
        o_num = 35
        self.sam_set.update_bounds(o_num)
        nptest.assert_array_equal(self.sam_set._left,
                                  np.repeat([self.domain[:, 0]], o_num, 0))
        nptest.assert_array_equal(self.sam_set._right,
                                  np.repeat([self.domain[:, 1]], o_num, 0))
        nptest.assert_array_equal(self.sam_set._width,
                                  np.repeat([self.domain[:, 1] - self.domain[:, 0]], o_num, 0))

    def test_update_bounds_local(self):
        """
        Check update_bounds_local
        """
        self.sam_set.global_to_local()
        self.sam_set.set_domain(self.domain)
        self.sam_set.update_bounds_local()
        local_size = self.sam_set.get_values_local().shape[0]
        nptest.assert_array_equal(self.sam_set._left_local,
                                  np.repeat([self.domain[:, 0]], local_size, 0))
        nptest.assert_array_equal(self.sam_set._right_local,
                                  np.repeat([self.domain[:, 1]], local_size, 0))
        nptest.assert_array_equal(self.sam_set._width_local,
                                  np.repeat([self.domain[:, 1] - self.domain[:, 0]], local_size,
                                            0))
        o_num = 35
        self.sam_set.update_bounds_local(o_num)
        nptest.assert_array_equal(self.sam_set._left_local,
                                  np.repeat([self.domain[:, 0]], o_num, 0))
        nptest.assert_array_equal(self.sam_set._right_local,
                                  np.repeat([self.domain[:, 1]], o_num, 0))
        nptest.assert_array_equal(self.sam_set._width_local,
                                  np.repeat([self.domain[:, 1] - self.domain[:, 0]], o_num, 0))

    def test_check_dim(self):
        """
        Check set_dim.
        """
        self.assertEqual(self.dim, self.sam_set.get_dim())

    def test_set_values(self):
        """
        Check set_values.
        """
        values = np.ones((150, self.dim))
        self.sam_set.set_values(values)
        nptest.assert_array_equal(util.fix_dimensions_data(values),
                                  self.sam_set.get_values())

    def test_set_values_local(self):
        """
        Check set_values_local.
        """
        values = np.ones((15, self.dim))
        self.sam_set.set_values_local(values)
        nptest.assert_array_equal(util.fix_dimensions_data(values),
                                  self.sam_set.get_values_local())

    def test_get_values(self):
        """
        Check get_samples.
        """
        nptest.assert_array_equal(util.fix_dimensions_data(self.values),
                                  self.sam_set.get_values())

    def test_get_shape(self):
        """
        Check get_samples.
        """
        nptest.assert_array_equal(util.fix_dimensions_data(self.values).shape,
                                  self.sam_set.shape())

    def test_append_values(self):
        """
        Check appending of values.
        """
        new_values = np.zeros((10, self.dim))
        self.sam_set.append_values(new_values)
        nptest.assert_array_equal(util.fix_dimensions_data(new_values),
                                  self.sam_set.get_values()[self.num::, :])

    def test_append_values_local(self):
        """
        Check appending of local values.
        """
        new_values = np.zeros((10, self.dim))
        self.sam_set.global_to_local()

        local_size = self.sam_set.get_values_local().shape[0]
        self.sam_set.append_values_local(new_values)
        nptest.assert_array_equal(util.fix_dimensions_data(new_values),
                                  self.sam_set.get_values_local()[local_size::, :])

    def test_get_dim(self):
        """
        Check to see if dimensions are correct.
        """
        self.assertEqual(self.dim, self.sam_set.get_dim())

    def test_probabilities(self):
        """
        Check probability methods.
        """
        prob = 1.0 / float(self.num) * np.ones((self.num,))
        self.sam_set.set_probabilities(prob)
        self.sam_set.check_num()
        nptest.assert_array_equal(prob, self.sam_set.get_probabilities())

    def test_densities(self):
        """
        Check density methods.
        """
        prob = 1.0 / float(self.num) * np.ones((self.num,))
        self.sam_set.set_probabilities(prob)
        self.sam_set.estimate_volume_mc()
        self.sam_set.set_densities()
        self.sam_set.check_num()
        vol = self.sam_set.get_volumes()
        nptest.assert_array_equal(prob / vol, self.sam_set.get_densities())
        den = np.ones((self.num,))
        self.sam_set.set_densities(den)
        nptest.assert_array_equal(den, self.sam_set.get_densities())

    def test_volumes(self):
        """
        Check volume methods.
        """
        vol = 1.0 / float(self.num) * np.ones((self.num,))
        self.sam_set.set_volumes(vol)
        self.sam_set.check_num()
        nptest.assert_array_equal(vol, self.sam_set.get_volumes())

    def test_error_estimates(self):
        """
        Check error estimate methods.
        """
        ee = np.ones((self.num, self.dim))
        self.sam_set.set_error_estimates(ee)
        self.sam_set.check_num()
        nptest.assert_array_equal(ee, self.sam_set.get_error_estimates())

    def test_region(self):
        """
        Check region methods.
        """
        region = np.ones((self.num,), dtype=np.int)
        self.sam_set.set_region(region)
        self.sam_set.check_num()
        nptest.assert_array_equal(region, self.sam_set.get_region())

    def test_error_id(self):
        """
        Check error identifier methods.
        """
        error_id = np.ones((self.num,))
        self.sam_set.set_error_id(error_id)
        self.sam_set.check_num()
        nptest.assert_array_equal(error_id, self.sam_set.get_error_id())

    def test_jacobian_methods(self):
        """
        Check jacobian methods.
        """
        jac = np.ones((self.num, 3, self.dim))
        self.sam_set.set_jacobians(jac)
        self.sam_set.check_num()
        nptest.assert_array_equal(jac, self.sam_set.get_jacobians())

    def test_check_num(self):
        """
        Check check_num.
        """
        prob = 1.0 / float(self.num) * np.ones((self.num,))
        self.sam_set.set_probabilities(prob)
        vol = 1.0 / float(self.num) * np.ones((self.num,))
        self.sam_set.set_volumes(vol)
        ee = np.ones((self.num, self.dim))
        self.sam_set.set_error_estimates(ee)
        jac = np.ones((self.num, 3, self.dim))
        self.sam_set.set_jacobians(jac)
        num = self.sam_set.check_num()
        self.assertEqual(self.num, num)
        new_values = np.zeros((10, self.dim))
        self.sam_set.append_values(new_values)
        self.assertRaises(sample.length_not_matching, self.sam_set.check_num)

    def test_kd_tree(self):
        """
        Check features of the KD Tree
        """
        self.sam_set.set_kdtree()
        self.sam_set.get_kdtree()

    def test_parallel_features(self):
        """
        Check parallel features.
        """
        prob = 1.0 / float(self.num) * np.ones((self.num,))
        self.sam_set.set_probabilities(prob)
        vol = 1.0 / float(self.num) * np.ones((self.num,))
        self.sam_set.set_volumes(vol)
        ee = np.ones((self.num, self.dim))
        self.sam_set.set_error_estimates(ee)
        jac = np.ones((self.num, 3, self.dim))
        self.sam_set.set_jacobians(jac)
        self.sam_set.global_to_local()
        self.assertNotIn(None, self.sam_set._values_local)
        if comm.size > 1:
            for array_name in sample.sample_set.array_names:
                current_array = getattr(self.sam_set, array_name + "_local")
                if current_array is not None:
                    self.assertGreater(getattr(self.sam_set,
                                               array_name).shape[0], current_array.shape[0])
                    local_size = current_array.shape[0]
                    num = comm.allreduce(local_size, op=MPI.SUM)
                    self.assertEqual(num, self.num)
                    current_array_global = util.get_global_values(
                        current_array)
                    nptest.assert_array_equal(getattr(self.sam_set,
                                                      array_name), current_array_global)
                    if array_name is "_values":
                        assert self.sam_set.shape_local() == (local_size,
                                                              self.dim)
        else:
            for array_name in sample.sample_set.array_names:
                current_array = getattr(self.sam_set, array_name + "_local")
                if current_array is not None:
                    nptest.assert_array_equal(getattr(self.sam_set,
                                                      array_name), current_array)
                    if array_name is "_values":
                        assert self.sam_set.shape_local() == (self.num,
                                                              self.dim)

        for array_name in sample.sample_set.array_names:
            current_array = getattr(self.sam_set, array_name)
            if current_array is not None:
                setattr(self.sam_set, array_name + "_old", current_array)
                current_array = None
        self.sam_set.local_to_global()
        for array_name in sample.sample_set.array_names:
            current_array = getattr(self.sam_set, array_name + "_local")
            if current_array is not None:
                nptest.assert_array_equal(getattr(self.sam_set, array_name),
                                          getattr(self.sam_set, array_name +
                                                  "_old"))

    def test_domain(self):
        """
        Test domain information.
        """
        domain = np.ones((self.dim, 2))
        self.sam_set.set_domain(domain)
        nptest.assert_array_equal(domain, self.sam_set.get_domain())


class Test_sample_set_1d(Test_sample_set):
    def setUp(self):
        self.dim = 1
        self.num = 100
        self.values = np.ones((self.num, self.dim))
        self.sam_set = sample.sample_set(dim=self.dim)
        self.sam_set.set_values(self.values)
        self.domain = np.array([[0, 1]], dtype=np.float)


class Test_discretization_simple(unittest.TestCase):
    def setUp(self):
        self.dim1 = 3
        self.num = 100
        self.dim2 = 1
        values1 = np.ones((self.num, self.dim1))
        values2 = np.ones((self.num, self.dim2))
        values3 = np.ones((self.num, self.dim2))
        self.input_set = sample.sample_set(dim=self.dim1)
        self.output_set = sample.sample_set(dim=self.dim2)
        self.output_probability_set = sample.sample_set(dim=self.dim2)
        self.input_set.set_values(values1)
        self.output_set.set_values(values2)
        self.output_probability_set.set_values(values3)
        self.disc = sample.discretization(input_sample_set=self.input_set,
                                          output_sample_set=self.output_set,
                                          output_probability_set=self.output_probability_set)

    def test_check_nums(self):
        """
        Test number checking.
        """
        num = self.disc.check_nums()
        self.assertEqual(num, self.num)

    def test_clip(self):
        """
        Test clipping of discretization.
        """
        cnum = int(0.5 * self.num)
        disc_clipped = self.disc.clip(cnum)
        nptest.assert_array_equal(self.disc._input_sample_set._values[0:cnum, :],
                                  disc_clipped._input_sample_set._values)
        nptest.assert_array_equal(self.disc._output_sample_set._values[0:cnum, :],
                                  disc_clipped._output_sample_set._values)

    def test_slicing(self):
        """
        Test `bet.sample.discretization.choose_inputs_outputs`
        """
        self.disc._output_sample_set.set_error_estimates(
            np.ones((self.num, self.dim2)))
        self.disc._input_sample_set.set_jacobians(
            np.ones((self.num, self.dim2, self.dim1)))

        disc_new = self.disc.choose_inputs_outputs(inputs=[0, 2], outputs=[0])
        nptest.assert_array_equal(self.disc._input_sample_set._values[:, [0, 2]],
                                  disc_new._input_sample_set._values)
        nptest.assert_array_equal(self.disc._output_sample_set._values[:, [0]],
                                  disc_new._output_sample_set._values)
        nptest.assert_array_equal(self.disc._output_sample_set._error_estimates[:, [0]],
                                  disc_new._output_sample_set._error_estimates)
        self.assertEqual(
            disc_new._input_sample_set._jacobians.shape, (self.num, 1, 2))

    def test_set_io_ptr(self):
        """
        Test setting io ptr
        """
        # TODO be careful if we change Kdtree
        self.disc.set_io_ptr(globalize=True)
        self.disc.get_io_ptr()
        self.disc.set_io_ptr(globalize=False)
        self.disc.get_io_ptr()
        self.disc.globalize_ptrs()

    def test_set_emulated_ii_ptr(self):
        """
        Test setting emulated ii ptr
        """
        # TODO be careful if we change Kdtree
        values = np.ones((10, self.dim1))
        self.emulated = sample.sample_set(dim=self.dim1)
        self.emulated.set_values(values)
        self.disc._emulated_input_sample_set = self.emulated
        self.disc.set_emulated_ii_ptr(globalize=True)
        self.disc.get_emulated_ii_ptr()
        self.disc.set_emulated_ii_ptr(globalize=False)
        self.disc._emulated_input_sample_set.local_to_global()
        self.disc.get_emulated_ii_ptr()
        self.disc.globalize_ptrs()

    def test_set_emulated_oo_ptr(self):
        """
        Test setting emulated oo ptr
        """
        # TODO be careful if we change Kdtree
        values = np.ones((3, self.dim2))
        self.emulated = sample.sample_set(dim=self.dim2)
        self.emulated.set_values(values)
        self.disc._emulated_output_sample_set = self.emulated
        self.disc.set_emulated_oo_ptr(globalize=True)
        self.disc.get_emulated_oo_ptr()
        self.disc.set_emulated_oo_ptr(globalize=False)
        self.disc.get_emulated_oo_ptr()
        self.disc.globalize_ptrs()

    def test_set_input_sample_set(self):
        """
        Test setting input sample set
        """
        test_set = sample.sample_set(dim=self.dim1)
        self.disc.set_input_sample_set(test_set)

    def test_get_input_sample_set(self):
        """
        Test getting input sample set
        """
        self.disc.get_input_sample_set()

    def test_set_emulated_input_sample_set(self):
        """
        Test setting emulated input sample set
        """
        test_set = sample.sample_set(dim=self.dim1)
        self.disc.set_emulated_input_sample_set(test_set)

    def test_get_emulated_input_sample_set(self):
        """
        Test getting emulated input sample set
        """
        self.disc.get_emulated_input_sample_set()

    def test_set_output_sample_set(self):
        """
        Test setting output sample set
        """
        test_set = sample.sample_set(dim=self.dim2)
        self.disc.set_output_sample_set(test_set)

    def test_get_output_sample_set(self):
        """
        Test getting output sample set
        """
        self.disc.get_output_sample_set()

    def test_set_output_probability_set(self):
        """
        Test setting output probability sample set
        """
        test_set = sample.sample_set(dim=self.dim2)
        self.disc.set_output_probability_set(test_set)

    def test_get_output_probability_set(self):
        """
        Test getting output probability sample set
        """
        self.disc.get_output_probability_set()

    def test_set_emulated_output_sample_set(self):
        """
        Test setting emulated output sample set
        """
        test_set = sample.sample_set(dim=self.dim2)
        self.disc.set_emulated_output_sample_set(test_set)

    def test_get_emulated_output_sample_set(self):
        """
        Test getting emulated output sample set
        """
        self.disc.get_emulated_output_sample_set()

    def test_save_load_discretization(self):
        """
        Test saving and loading of discretization
        """
        file_name = os.path.join(local_path, 'testfile')
        globalize = True
        util.save_object(self.disc, file_name, globalize)
        comm.barrier()
        if comm.size > 1 and not globalize:
            local_file_name = os.path.os.path.join(os.path.dirname(file_name),
                                                   "proc{}_{}".format(comm.rank, os.path.basename(file_name)))
        else:
            local_file_name = file_name

        loaded_disc = util.load_object(local_file_name)

        comm.barrier()
        assert loaded_disc == self.disc


        if comm.rank == 0 and globalize:
            os.remove(local_file_name+'.p')
        elif not globalize:
            os.remove(local_file_name+'.p')
        globalize = False
        util.save_object(self.disc, file_name, globalize)
        comm.barrier()
        if comm.size > 1 and not globalize:
            local_file_name = os.path.os.path.join(os.path.dirname(file_name),
                                                   "proc{}_{}".format(comm.rank,
                                                                      os.path.basename(file_name)))
        else:
            local_file_name = file_name

        loaded_disc = util.load_object(local_file_name)

        comm.barrier()
        assert loaded_disc == self.disc

        if comm.rank == 0 and globalize:
            os.remove(file_name+'.p')
        else:
            os.remove(local_file_name+'.p')


    def test_copy_discretization(self):
        """
        Test copying of discretization
        """
        copied_disc = self.disc.copy()

        assert copied_disc == self.disc

    def test_estimate_input_volume_emulated(self):
        """

        Testing :meth:`bet.discretization.estimate_input_volume_emulated`

        """
        lam_left = np.array([0.0, .25, .4])
        lam_right = np.array([1.0, 4.0, .5])
        lam_width = lam_right - lam_left

        lam_domain = np.zeros((3, 2))
        lam_domain[:, 0] = lam_left
        lam_domain[:, 1] = lam_right

        num_samples_dim = 2
        start = lam_left + lam_width / (2 * num_samples_dim)
        stop = lam_right - lam_width / (2 * num_samples_dim)
        d1_arrays = []

        for l, r in zip(start, stop):
            d1_arrays.append(np.linspace(l, r, num_samples_dim))

        s_set = sample.sample_set(util.meshgrid_ndim(d1_arrays).shape[1])
        s_set.set_domain(lam_domain)
        s_set.set_values(util.meshgrid_ndim(d1_arrays))

        volume_exact = 1.0 / s_set._values.shape[0]

        emulated_samples = s_set.copy()
        emulated_samples.update_bounds_local(1001)
        emulated_samples.set_values_local(emulated_samples._width_local
                                          * np.random.random((1001, emulated_samples.get_dim())) +
                                          emulated_samples._left_local)

        self.disc.set_input_sample_set(s_set)
        self.disc.set_emulated_input_sample_set(emulated_samples)
        self.disc.estimate_input_volume_emulated()

        lam_vol = self.disc._input_sample_set._volumes

        # Check the dimension.
        nptest.assert_array_equal(lam_vol.shape, (len(s_set._values), ))

        # Check that the volumes are within a tolerance for a regular grid of
        # samples.
        nptest.assert_array_almost_equal(lam_vol, volume_exact, 1)
        nptest.assert_almost_equal(np.sum(lam_vol), 1.0)

    def test_estimate_output_volume_emulated(self):
        """

        Testing :meth:`bet.discretization.estimate_output_volume_emulated`

        """
        lam_left = np.array([0.0])
        lam_right = np.array([1.0])
        lam_width = lam_right - lam_left

        lam_domain = np.zeros((1, 2))
        lam_domain[:, 0] = lam_left
        lam_domain[:, 1] = lam_right

        num_samples_dim = 2
        start = lam_left + lam_width / (2 * num_samples_dim)
        stop = lam_right - lam_width / (2 * num_samples_dim)
        d1_arrays = []

        for l, r in zip(start, stop):
            d1_arrays.append(np.linspace(l, r, num_samples_dim))

        s_set = sample.sample_set(util.meshgrid_ndim(d1_arrays).shape[1])
        s_set.set_domain(lam_domain)
        s_set.set_values(util.meshgrid_ndim(d1_arrays))

        volume_exact = 1.0 / s_set._values.shape[0]

        emulated_samples = s_set.copy()
        emulated_samples.update_bounds_local(1001)
        emulated_samples.set_values_local(emulated_samples._width_local
                                          * np.random.random((1001, emulated_samples.get_dim())) +
                                          emulated_samples._left_local)

        self.disc.set_output_sample_set(s_set)
        self.disc.set_emulated_output_sample_set(emulated_samples)
        self.disc.estimate_output_volume_emulated()

        lam_vol = self.disc._output_sample_set._volumes

        # Check the dimension.
        nptest.assert_array_equal(lam_vol.shape, (len(s_set._values), ))

        # Check that the volumes are within a tolerance for a regular grid of
        # samples.
        nptest.assert_array_almost_equal(lam_vol, volume_exact, 1)
        nptest.assert_almost_equal(np.sum(lam_vol), 1.0)


class TestEstimateVolume(unittest.TestCase):
    """
    Test :meth:`bet.calculateP.calculateP.estimate_volulme`.
    """

    def setUp(self):
        """
        Test dimension, number of samples, and that all the samples are within
        lambda_domain.
        """
        lam_left = np.array([0.0, .25, .4])
        lam_right = np.array([1.0, 4.0, .5])
        lam_width = lam_right - lam_left

        self.lam_domain = np.zeros((3, 2))
        self.lam_domain[:, 0] = lam_left
        self.lam_domain[:, 1] = lam_right

        num_samples_dim = 2
        start = lam_left + lam_width / (2 * num_samples_dim)
        stop = lam_right - lam_width / (2 * num_samples_dim)
        d1_arrays = []

        for l, r in zip(start, stop):
            d1_arrays.append(np.linspace(l, r, num_samples_dim))

        self.s_set = sample.sample_set(util.meshgrid_ndim(d1_arrays).shape[1])
        self.s_set.set_domain(self.lam_domain)
        self.s_set.set_values(util.meshgrid_ndim(d1_arrays))
        print(util.meshgrid_ndim(d1_arrays).shape)
        self.volume_exact = 1.0 / self.s_set._values.shape[0]
        self.s_set.estimate_volume(n_mc_points=1001)
        self.lam_vol = self.s_set._volumes

    def test_dimension(self):
        """
        Check the dimension.
        """
        print(self.lam_vol.shape, self.s_set._values.shape)
        nptest.assert_array_equal(
            self.lam_vol.shape, (len(self.s_set._values), ))

    def test_volumes(self):
        """
        Check that the volumes are within a tolerance for a regular grid of
        samples.
        """
        nptest.assert_array_almost_equal(self.lam_vol, self.volume_exact, 1)
        nptest.assert_almost_equal(np.sum(self.lam_vol), 1.0)


class TestEstimateVolumeEmulated(unittest.TestCase):
    """
    Test :meth:`bet.calculateP.calculateP.estimate_volulme_emulated`.
    """

    def setUp(self):
        """
        Test dimension, number of samples, and that all the samples are within
        lambda_domain.
        """
        lam_left = np.array([0.0, .25, .4])
        lam_right = np.array([1.0, 4.0, .5])
        lam_width = lam_right - lam_left

        self.lam_domain = np.zeros((3, 2))
        self.lam_domain[:, 0] = lam_left
        self.lam_domain[:, 1] = lam_right

        num_samples_dim = 2
        start = lam_left + lam_width / (2 * num_samples_dim)
        stop = lam_right - lam_width / (2 * num_samples_dim)
        d1_arrays = []

        for l, r in zip(start, stop):
            d1_arrays.append(np.linspace(l, r, num_samples_dim))

        self.s_set = sample.sample_set(util.meshgrid_ndim(d1_arrays).shape[1])
        self.s_set.set_domain(self.lam_domain)
        self.s_set.set_values(util.meshgrid_ndim(d1_arrays))
        print(util.meshgrid_ndim(d1_arrays).shape)
        self.volume_exact = 1.0 / self.s_set._values.shape[0]
        emulated_samples = self.s_set.copy()
        emulated_samples.update_bounds_local(1001)
        emulated_samples.set_values_local(emulated_samples._width_local
                                          * np.random.random((1001, emulated_samples.get_dim())) +
                                          emulated_samples._left_local)
        self.s_set.estimate_volume_emulated(emulated_samples)
        self.lam_vol = self.s_set._volumes

    def test_dimension(self):
        """
        Check the dimension.
        """
        print(self.lam_vol.shape, self.s_set._values.shape)
        nptest.assert_array_equal(
            self.lam_vol.shape, (len(self.s_set._values), ))

    def test_volumes(self):
        """
        Check that the volumes are within a tolerance for a regular grid of
        samples.
        """
        nptest.assert_array_almost_equal(self.lam_vol, self.volume_exact, 1)
        nptest.assert_almost_equal(np.sum(self.lam_vol), 1.0)


class TestEstimateLocalVolume(unittest.TestCase):
    """
    Test :meth:`bet.calculateP.calculateP.estimate_local_volulme`.
    """

    def setUp(self):
        """
        Test dimension, number of samples, and that all the samples are within
        lambda_domain.

        """
        lam_left = np.array([0.0, .25, .4])
        lam_right = np.array([1.0, 4.0, .5])
        lam_width = lam_right - lam_left

        self.lam_domain = np.zeros((3, 2))
        self.lam_domain[:, 0] = lam_left
        self.lam_domain[:, 1] = lam_right

        num_samples_dim = 2
        start = lam_left + lam_width / (2 * num_samples_dim)
        stop = lam_right - lam_width / (2 * num_samples_dim)
        d1_arrays = []

        for l, r in zip(start, stop):
            d1_arrays.append(np.linspace(l, r, num_samples_dim))

        self.s_set = sample.sample_set(util.meshgrid_ndim(d1_arrays).shape[1])
        self.s_set.set_domain(self.lam_domain)
        self.s_set.set_values(util.meshgrid_ndim(d1_arrays))
        self.volume_exact = 1.0 / self.s_set._values.shape[0]
        self.s_set.estimate_local_volume()
        self.lam_vol = self.s_set._volumes

    def test_dimension(self):
        """
        Check the dimension.
        """
        nptest.assert_array_equal(
            self.lam_vol.shape, (len(self.s_set._values), ))

    def test_volumes(self):
        """
        Check that the volumes are within a tolerance for a regular grid of
        samples.
        """
        nptest.assert_array_almost_equal(self.lam_vol, self.volume_exact, 2)
        nptest.assert_almost_equal(np.sum(self.lam_vol), 1.0)


class TestExactVolume1D(unittest.TestCase):
    """
    Test :meth:`bet.calculateP.calculateP.exact_volume_1D`.
    """

    def setUp(self):
        """
        Test dimension, number of samples, and that all the samples are within
        lambda_domain.
        """
        num_samples = 10
        self.lam_domain = np.array([[.0, .1]])
        edges = np.linspace(self.lam_domain[:, 0], self.lam_domain[:, 1],
                            num_samples + 1)
        self.samples = (edges[1:] + edges[:-1]) * .5
        np.random.shuffle(self.samples)
        self.volume_exact = 1. / self.samples.shape[0]
        self.volume_exact = self.volume_exact * np.ones((num_samples,))
        s_set = sample.voronoi_sample_set(dim=1)
        s_set.set_domain(self.lam_domain)
        s_set.set_values(self.samples)
        s_set.exact_volume_1D()
        self.lam_vol = s_set.get_volumes()

    def test_dimension(self):
        """
        Check the dimension.
        """
        nptest.assert_array_equal(self.lam_vol.shape, (len(self.samples), ))

    def test_volumes(self):
        """
        Check that the volumes are within a tolerance for a regular grid of
        samples.
        """
        nptest.assert_array_almost_equal(self.lam_vol, self.volume_exact)
        nptest.assert_almost_equal(np.sum(self.lam_vol), 1.0)


class TestExactVolume2D(unittest.TestCase):
    """
    Test :meth:`bet.calculateP.calculateP.exact_volume_2D`.
    """

    def setUp(self):
        """
        Test dimension, number of samples, and that all the samples are within
        lambda_domain.
        """
        sampler = bsam.sampler(None)
        self.input_samples = sample.sample_set(2)
        self.input_samples.set_domain(np.array([[0.0, 1.0], [0.0, 1.0]]))
        self.input_samples = sampler.regular_sample_set(
            self.input_samples, num_samples_per_dim=[10, 9])
        self.input_samples.exact_volume_2D()
        self.vol1 = np.copy(self.input_samples._volumes)
        self.input_samples.estimate_volume_mc()
        self.vol2 = np.copy(self.input_samples._volumes)

    def test_volumes(self):
        """
        Check that the volumes are within a tolerance for a regular grid of
        samples.
        """
        nptest.assert_array_almost_equal(self.vol1, self.vol2)
        nptest.assert_almost_equal(np.sum(self.vol1), 1.0)


class TestEstimateRadii(unittest.TestCase):
    """
    Test :meth:`bet.calculateP.calculateP.estimate_radii`.
    """

    def setUp(self):
        """
        Test dimension, number of samples, and that all the samples are within
        lambda_domain.

        """
        lam_left = np.array([0.0, 0.5, 0.5])
        lam_right = np.array([1.0, 1.5, 1.5])
        lam_width = lam_right - lam_left

        self.lam_domain = np.zeros((3, 2))
        self.lam_domain[:, 0] = lam_left
        self.lam_domain[:, 1] = lam_right

        num_samples_dim = 2
        start = lam_left + lam_width / (2 * num_samples_dim)
        stop = lam_right - lam_width / (2 * num_samples_dim)
        d1_arrays = []

        for l, r in zip(start, stop):
            d1_arrays.append(np.linspace(l, r, num_samples_dim))

        self.s_set = sample.sample_set(util.meshgrid_ndim(d1_arrays).shape[1])
        self.s_set.set_domain(self.lam_domain)
        self.s_set.set_values(util.meshgrid_ndim(d1_arrays))

        self.radii_exact = np.sqrt(3 * .25**2)

        self.s_set.estimate_radii(normalize=False)
        self.s_set.estimate_radii()
        self.rad = self.s_set._radii
        self.norm_rad = self.s_set._normalized_radii

    def test_dimension(self):
        """
        Check the dimension.
        """
        nptest.assert_array_equal(self.rad.shape, (len(self.s_set._values), ))
        nptest.assert_array_equal(
            self.norm_rad.shape, (len(self.s_set._values), ))

    def test_radii(self):
        """
        Check that the radii are within a tolerance for a regular grid of
        samples.
        """
        nptest.assert_array_almost_equal(self.rad, self.radii_exact, 1)
        nptest.assert_array_almost_equal(self.norm_rad, self.radii_exact, 1)


class TestEstimateRadiiAndVolume(unittest.TestCase):
    """
    Test :meth:`bet.calculateP.calculateP.estimate_radii_and_volume`.
    """

    def setUp(self):
        """
        Test dimension, number of samples, and that all the samples are within
        lambda_domain.

        """
        lam_left = np.array([0.0, 0.5, 0.5])
        lam_right = np.array([1.0, 1.5, 1.5])
        lam_width = lam_right - lam_left

        self.lam_domain = np.zeros((3, 2))
        self.lam_domain[:, 0] = lam_left
        self.lam_domain[:, 1] = lam_right

        num_samples_dim = 2
        start = lam_left + lam_width / (2 * num_samples_dim)
        stop = lam_right - lam_width / (2 * num_samples_dim)
        d1_arrays = []

        for l, r in zip(start, stop):
            d1_arrays.append(np.linspace(l, r, num_samples_dim))

        self.s_set = sample.sample_set(util.meshgrid_ndim(d1_arrays).shape[1])
        self.s_set.set_domain(self.lam_domain)
        self.s_set.set_values(util.meshgrid_ndim(d1_arrays))
        self.volume_exact = 1.0 / self.s_set._values.shape[0]

        self.radii_exact = np.sqrt(3 * .25**2)

        self.s_set.estimate_radii_and_volume(normalize=False)
        self.s_set.estimate_radii_and_volume()
        self.lam_vol = self.s_set._volumes
        self.rad = self.s_set._radii
        self.norm_rad = self.s_set._normalized_radii

    def test_dimension(self):
        """
        Check the dimension.
        """
        nptest.assert_array_equal(self.rad.shape, (len(self.s_set._values), ))
        nptest.assert_array_equal(
            self.norm_rad.shape, (len(self.s_set._values), ))
        nptest.assert_array_equal(
            self.lam_vol.shape, (len(self.s_set._values), ))

    def test_radii(self):
        """
        Check that the volumes are within a tolerance for a regular grid of
        samples.
        """
        nptest.assert_array_almost_equal(self.rad, self.radii_exact, 1)
        nptest.assert_array_almost_equal(self.norm_rad, self.radii_exact, 1)
        nptest.assert_array_almost_equal(self.lam_vol, self.volume_exact, 1)
        nptest.assert_almost_equal(np.sum(self.lam_vol), 1.0)


class Test_rectangle_sample_set(unittest.TestCase):
    def setUp(self):
        self.dim = 2
        self.sam_set = sample.rectangle_sample_set(dim=self.dim)
        # maximum number
        nprocs = 8
        self.nprocs = nprocs
        n = np.linspace(0.1, 0.9, nprocs)
        maxes = [[n[i], n[i]] for i in range(1, nprocs)]
        mins = [[n[i], n[i]] for i in range(nprocs - 1)]
        self.sam_set.setup(maxes, mins)
        self.domain = np.array([[0, 1], [0, 1]], dtype=np.float)
        self.sam_set.set_domain(self.domain)
        self.num = self.sam_set.check_num()

    def test_query(self):
        """
        Check querying
        """
        n = np.linspace(0.1, 0.9, self.nprocs)
        x = np.array([[n[i] + 1E-5, n[i] + 1E-5]
                      for i in range(self.nprocs - 1)])
        (d, ptr) = self.sam_set.query(x)
        nptest.assert_array_equal(ptr, np.arange(self.nprocs - 1))

    def test_volumes(self):
        """
        Check volume calculation
        """
        self.sam_set.exact_volume_lebesgue()
        volumes = self.sam_set.get_volumes()
        volumes_exact = (0.8 / (self.nprocs - 1))**2
        total_vol_exact = 1 - (self.nprocs - 1) * volumes_exact
        nptest.assert_array_almost_equal(volumes[:-1], volumes_exact)
        nptest.assert_array_almost_equal(volumes[-1], total_vol_exact)


class Test_ball_sample_set(unittest.TestCase):
    def setUp(self):
        self.dim = 2
        self.sam_set = sample.ball_sample_set(dim=self.dim)
        # max number of processors supported in test.
        nprocs = 8
        self.nprocs = nprocs
        n = np.linspace(0.1, 0.9, nprocs)
        centers = [[n[i], n[i]] for i in range(nprocs - 1)]
        radii = [0.1] * (nprocs - 1)
        self.sam_set.setup(centers, radii)
        self.domain = np.array([[0, 1], [0, 1]], dtype=np.float)
        self.sam_set.set_domain(self.domain)
        self.num = self.sam_set.check_num()

    def test_query(self):
        """
        Check querying
        """
        n = np.linspace(0.1, 0.9, self.nprocs)
        x = np.array([[n[i] + 1E-5, n[i] + 1E-5]
                      for i in range(self.nprocs - 1)])
        (d, ptr) = self.sam_set.query(x)
        nptest.assert_array_equal(ptr, np.arange(self.nprocs - 1))

    def test_volumes(self):
        """
        Check volume calculation
        """
        self.sam_set.exact_volume()
        volumes = self.sam_set.get_volumes()
        nptest.assert_array_almost_equal(volumes[:-1], np.pi / 100)
        leftover_volume = 1 - (self.nprocs - 1) * np.pi / 100
        nptest.assert_array_almost_equal(volumes[-1], leftover_volume)


class Test_cartesian_sample_set(unittest.TestCase):
    def setUp(self):
        self.dim = 2
        self.sam_set = sample.cartesian_sample_set(dim=self.dim)
        # number of processors test should support (for dim<=2 only)
        nprocs = 10
        self.nprocs = int(np.ceil(np.sqrt(nprocs))**2)
        # equispaced grid in each dimension
        vi = np.linspace(0, 1, 1 + int(np.sqrt(self.nprocs)))
        xi = [vi, vi]
        self.sam_set.setup(xi)
        self.domain = np.array([[0, 1], [0, 1]], dtype=np.float)
        self.sam_set.set_domain(self.domain)
        self.num = self.sam_set.check_num()

    def test_query(self):
        """
        Check querying
        """
        x = self.sam_set.get_values()[:-1, :] + 1E-5
        (d, ptr) = self.sam_set.query(x)
        nptest.assert_array_equal(ptr, np.arange(self.nprocs))

    def test_volumes(self):
        """
        Check volume calculation
        """
        self.sam_set.exact_volume_lebesgue()
        volumes = self.sam_set.get_volumes()
        nptest.assert_array_almost_equal(volumes[:-1], 1. / self.nprocs)
        assert volumes[-1] == 0



# Copyright (C) 2014-2020 The BET Development Team


import bet.sampling.basicSampling as bsam
import bet.postProcess.compareP as compP

"""
Compare marginals of two probability measures based on random variables with certain properties.
"""

# Initialize two sample sets
set1 = bsam.random_sample_set(rv=[['beta', {'loc': 0, 'scale': 2, 'a': 3, 'b': 2}],
                              ['beta', {'loc': 0, 'scale': 2, 'a': 3, 'b': 2}]],
                              input_obj=2, num_samples=300)
set2 = bsam.random_sample_set(rv=[['beta', {'loc': 0, 'scale': 2, 'a': 2, 'b': 3}],
                              ['beta', {'loc': 0, 'scale': 2, 'a': 2, 'b': 3}]],
                              input_obj=2, num_samples=300)

# Initialize metric
mm = compP.compare(set1, set2, set2_init=True, set1_init=True)
mm.set_compare_set()

# Test different distance metrics with discrete distances and by integrating with quadrature.
print(mm.distance('tv'))
print('Total Variation')
print(mm.distance_marginal(i=0, functional='tv', normalize=False))
print(mm.distance_marginal_quad(i=0, functional='tv'))

print('KL Divergence')
print(mm.distance_marginal(i=0, functional='kl', normalize=False))
print(mm.distance_marginal_quad(i=0, functional='kl'))

print('Hellinger Distance')
print(mm.distance_marginal(i=0, functional='hell', normalize=False))
print(mm.distance_marginal_quad(i=0, functional='hell'))

print('Euclidean Norm')
print(mm.distance_marginal(i=0, functional='2', normalize=False))
print(mm.distance_marginal_quad(i=0, functional='2'))



# Copyright (C) 2014-2020 The BET Development Team

# -*- coding: utf-8 -*-
r"""
The user should set the environment variables LAUNCHER_DIR
and LAUNCHER_PPN below.
"""

import os
import scipy.io as sio
import sys
import numpy as np


def lb_model(input_data):
    num_runs = input_data.shape[0]
    num_runs_dim = input_data.shape[1]

    # Setup the job file for Launcher.
    f = open('launcher_runs.txt', 'w')
    for i in range(0, num_runs):
        output_str = sys.executable + ' myModel_serial.py ' + repr(i) + ' '
        for j in range(0, num_runs_dim):
            output_str = output_str + repr(input_data[i, j]) + ' '
        output_str += '\n'
        f.write(output_str)
    f.close()
    os.environ["LAUNCHER_JOB_FILE"] = "launcher_runs.txt"

    # USER SETS THESE ENVIRONMENT VARIABLES
    os.environ["LAUNCHER_DIR"] = "DIRECTORY_TO_LAUNCHER_REPO"
    os.environ["LAUNCHER_PPN"] = "NUM_OF_PROCS_TO_USE"

    # Execute Launcher to start multiple serial runs of FEniCS
    os.system("bash /home/troy/Packages/launcher/paramrun")

    # Read in data from files and cleanup files.
    QoI_samples = np.zeros((num_runs, 2))
    for i in range(0, num_runs):
        io_file_name = 'QoI_sample' + repr(i)
        io_mdat = sio.loadmat(io_file_name)
        QoI_samples[i, :] = io_mdat['output']
        io_file_str = io_file_name + '.mat'
        os.remove(io_file_str)

    return QoI_samples



# Copyright (C) 2014-2020 The BET Development Team

"""
This subpackage contains

* :mod:`~bet.sampling.basicSampling` a general class and associated set of methods that sample spaces and solve models through an interface.
* :class:`~bet.sampling.basicSampling.sampler` requests data (QoI) at a specified set of parameter samples.
* :mod:`~bet.sampling.LpGeneralizedSamples` provides methods for sampling on balls in Lp spaces.
* :mod:`~bet.sampling.useLUQ` provides methods for interfacing with the LUQ package.
"""
__all__ = ['basicSampling', 'LpGeneralizedSamples', 'useLUQ']



# Copyright (C) 2014-2020 The BET Development Team

"""
This module provides methods for plotting probabilities.
"""

import copy
import math
import numpy as np
import scipy.stats as stats
import matplotlib
import matplotlib.pyplot as plt
#plt.rc('text', usetex=True)
#plt.rc('font', family='serif')
from bet.Comm import comm, MPI
import bet.sample as sample


class dim_not_matching(Exception):
    """
    Exception for when the dimension is inconsistent.
    """


class bad_object(Exception):
    """
    Exception for when the wrong type of object is used.
    """


class missing_attribute(Exception):
    """
    Exception for missing attribute.
    """


def calculate_1D_marginal_probs(sample_set, nbins=20):
    r"""
    This estimates every marginal of a Voronoi probability measure
    described by the probabilities within the sample_set object with histograms.
    If the sample_set object is a discretization object, we assume
    that the probabilities to be plotted are from the input space on the
    emulated samples (if they exist) or the samples.
    (``discretization._emulated_input_sample_set._probabilties_local`` or
    ``discretization._input_sample_set._probabilties_local``).

    This assumes that the user has already run
    :meth:`~bet.calculateP.calculateP.prob_emulated` or :meth:`~bet.calculateP.calculateP.prob`.

    :param sample_set: Object containing samples and probabilities
    :type sample_set: :class:`~bet.sample.sample_set_base` or
        :class:`~bet.sample.discretization`
    :param nbins: Number of bins in each direction.
    :type nbins: :int or :class:`~numpy.ndarray` of shape (ndim,)
    :rtype: tuple
    :returns: (bins, marginals)

    """
    if isinstance(sample_set, sample.discretization):
        if sample_set.get_emulated_input_sample_set() is not None:
            sample_obj = sample_set._emulated_input_sample_set
        else:
            sample_obj = sample_set.get_input_sample_set()
        if sample_obj is None:
            raise missing_attribute("Missing emulated_input_sample_set")
    elif isinstance(sample_set, sample.sample_set_base):
        sample_obj = sample_set
    else:
        raise bad_object("Improper sample object")

    # Check for local probabilities
    if sample_obj._probabilities_local is None:
        if sample_obj._probabilities is None:
            raise missing_attribute("Missing probabilities")
        else:
            sample_obj.global_to_local()

    # Make list of bins if only an integer is given
    if isinstance(nbins, int):
        nbins = nbins * np.ones(sample_obj.get_dim(), dtype=np.int)

    # Create bins
    bins = []
    for i in range(sample_obj.get_dim()):
        bins.append(np.linspace(sample_obj.get_domain()[i][0],
                                sample_obj.get_domain()[i][1],
                                nbins[i] + 1))

    # Calculate marginals
    marginals = {}
    for i in range(sample_obj.get_dim()):
        [marg, _] = np.histogram(sample_obj.get_values_local()[:, i],
                                 bins=bins[i], weights=sample_obj.get_probabilities_local())
        marg_temp = np.copy(marg)
        comm.Allreduce([marg, MPI.DOUBLE], [marg_temp, MPI.DOUBLE], op=MPI.SUM)
        marginals[i] = marg_temp

    return (bins, marginals)


def calculate_2D_marginal_probs(sample_set, nbins=20):
    """
    This calculates every pair of marginals (or joint in 2d case) of
    input probability measure defined on a rectangular grid for Voronoi probabilities using histograms..
    If the sample_set object is a discretization object, we assume
    that the probabilities to be plotted are from the input space on the
    emulated samples (if they exist) or samples
    (``discretization._emulated_input_sample_set._probabilties_local`` or
    ``discretization._input_sample_set._probabilties_local``).

    This assumes that the user has already run
    :meth:`~bet.calculateP.calculateP.prob_emulated` or :meth:`~bet.calculateP.calculateP.prob`.


    :param sample_set: Object containing samples and probabilities
    :type sample_set: :class:`~bet.sample.sample_set_base`
        or :class:`~bet.sample.discretization`
    :param nbins: Number of bins in each direction.
    :type nbins: int or :class:`~numpy.ndarray` of shape (ndim,)
    :rtype: tuple
    :returns: (bins, marginals)

    """
    if isinstance(sample_set, sample.discretization):
        if sample_set._emulated_input_sample_set is not None:
            sample_obj = sample_set._emulated_input_sample_set
        else:
            sample_obj = sample_set.get_input_sample_set()
        if sample_obj is None:
            raise missing_attribute("Missing input_sample_set")
    elif isinstance(sample_set, sample.sample_set_base):
        sample_obj = sample_set
    else:
        raise bad_object("Improper sample object")

    # Check for local probabilities
    if sample_obj._probabilities_local is None:
        if sample_obj._probabilities is None:
            raise missing_attribute("Missing probabilities")
        else:
            sample_obj.global_to_local()

    if sample_obj.get_dim() < 2:
        raise dim_not_matching("Incompatible dimensions of sample set"
                               " for plotting")

    # Make list of bins if only an integer is given
    if isinstance(nbins, int):
        nbins = nbins * np.ones(sample_obj.get_dim(), dtype=np.int)

    # Create bins
    bins = []
    for i in range(sample_obj.get_dim()):
        bins.append(np.linspace(sample_obj.get_domain()[i][0],
                                sample_obj.get_domain()[i][1],
                                nbins[i] + 1))

    # Calculate marginals
    marginals = {}
    for i in range(sample_obj.get_dim()):
        for j in range(i + 1, sample_obj.get_dim()):
            (marg, _) = np.histogramdd(sample_obj.get_values_local()[:, [i, j]],
                                       bins=[bins[i], bins[j]],
                                       weights=sample_obj.get_probabilities_local())
            marg = np.ascontiguousarray(marg)
            marg_temp = np.copy(marg)
            comm.Allreduce([marg, MPI.DOUBLE], [marg_temp, MPI.DOUBLE],
                           op=MPI.SUM)
            marginals[(i, j)] = marg_temp

    return (bins, marginals)


def plot_1D_marginal_probs(marginals, bins, sample_set,
                           filename="file", lam_ref=None, interactive=False,
                           lambda_label=None, file_extension=".png"):
    """
    This makes plots of every single marginal probability of
    input probability measure on a 1D  grid from histograms.
    If the sample_set object is a discretization object, we assume
    that the probabilities to be plotted are from the input space.
    Useful for visualizing solutions of measure-based inverse problems.

    .. note::

        Do not specify the file extension in the file name.

    :param marginals: 1D marginal probabilities
    :type marginals: dictionary with int as keys and :class:`~numpy.ndarray` of
        shape (nbins+1,) as values :param bins: Endpoints of bins used in
        calculating marginals
    :type bins: :class:`~numpy.ndarray` of shape (nbins+1,)
    :param sample_set: Object containing samples and probabilities
    :type sample_set: :class:`~bet.sample.sample_set_base`
        or :class:`~bet.sample.discretization`
    :param filename: Prefix for output files.
    :type filename: str
    :param lam_ref: True parameters.
    :type lam_ref: :class:`~numpy.ndarray` of shape (ndim,) or None
    :param interactive: Whether or not to display interactive plots.
    :type interactive: bool
    :param lambda_label: Label for each parameter for plots.
    :type lambda_label: list of length nbins of strings or None
    :param string file_extension: file extenstion

    """
    if isinstance(sample_set, sample.discretization):
        sample_obj = sample_set._input_sample_set
    elif isinstance(sample_set, sample.sample_set_base):
        sample_obj = sample_set
    else:
        raise bad_object("Improper sample object")

    if lam_ref is None:
        lam_ref = sample_obj._reference_value

    lam_domain = sample_obj.get_domain()

    if comm.rank == 0:
        index = sorted(copy.deepcopy(list(marginals.keys())))
        for i in index:
            x_range = np.linspace(lam_domain[i, 0], lam_domain[i, 1],
                                  len(bins[i]) - 1)
            fig = plt.figure(i)
            ax = fig.add_subplot(111)
            ax.plot(x_range, marginals[i] / (bins[i][1] - bins[i][0]))
            ax.set_ylim(
                [0, 1.05 * np.max(marginals[i] / (bins[i][1] - bins[i][0]))])
            if lam_ref is not None:
                ax.plot(lam_ref[i],
                        0.02 * 1.05 *
                        np.max(marginals[i] / (bins[i][1] - bins[i][0])),
                        'ko', markersize=15)
            if lambda_label is None:
                label1 = r'$\lambda_{' + str(i + 1) + '}$'
            else:
                label1 = lambda_label[i]
            ax.set_xlabel(label1, fontsize=30)
            ax.set_ylabel(r'PDF', fontsize=30)
            ax.tick_params(axis='both', which='major',
                           labelsize=20)
            plt.tight_layout()
            fig.savefig(filename + "_1D_" + str(i) + file_extension,
                        transparent=True)
            if interactive:
                plt.show()
            else:
                plt.close()
            plt.clf()
    comm.barrier()


def plot_2D_marginal_probs(marginals, bins, sample_set,
                           filename="file", lam_ref=None, plot_surface=False, interactive=False,
                           lambda_label=None, file_extension=".png"):
    """
    This makes plots of every pair of marginals (or joint in 2d case) of
    input probability measure on a rectangular grid from histograms.
    If the sample_set object is a discretization object, we assume
    that the probabilities to be plotted are from the input space.
    Useful for visualizing solutions of measure-based inverse problems.

    .. note::

        Do not specify the file extension in the file name.

    :param marginals: 2D marginal probabilities
    :type marginals: dictionary with tuples of 2 integers as keys and
        :class:`~numpy.ndarray` of shape (nbins+1,) as values
    :param bins: Endpoints of bins used in calculating marginals
    :type bins: :class:`~numpy.ndarray` of shape (nbins+1,2)
    :param sample_set: Object containing samples and probabilities
    :type sample_set: :class:`~bet.sample.sample_set_base`
        or :class:`~bet.sample.discretization`
    :param filename: Prefix for output files.
    :type filename: str
    :param lam_ref: True parameters.
    :type lam_ref: :class:`~numpy.ndarray` of shape (ndim,) or None
    :param interactive: Whether or not to display interactive plots.
    :type interactive: bool
    :param lambda_label: Label for each parameter for plots.
    :type lambda_label: list of length nbins of strings or None
    :param string file_extension: file extenstion

    """
    if isinstance(sample_set, sample.discretization):
        sample_obj = sample_set._input_sample_set
    elif isinstance(sample_set, sample.sample_set_base):
        sample_obj = sample_set
    else:
        raise bad_object("Improper sample object")

    if lam_ref is None:
        lam_ref = sample_obj._reference_value

    lam_domain = sample_obj.get_domain()

    from matplotlib import cm
    if plot_surface:
        from mpl_toolkits.mplot3d import Axes3D
        from matplotlib.ticker import LinearLocator, FormatStrFormatter
    if comm.rank == 0:
        pairs = sorted(copy.deepcopy(list(marginals.keys())))
        for k, (i, j) in enumerate(pairs):
            fig = plt.figure(k)
            ax = fig.add_subplot(111)
            boxSize = (bins[i][1] - bins[i][0]) * (bins[j][1] - bins[j][0])
            quadmesh = ax.imshow(marginals[(i, j)].transpose() / boxSize,
                                 interpolation='bicubic', cmap=cm.CMRmap_r,
                                 extent=[lam_domain[i][0], lam_domain[i][1],
                                         lam_domain[j][0], lam_domain[j][1]], origin='lower',
                                 vmax=marginals[(i, j)].max() / boxSize, vmin=0,
                                 aspect='auto')
            if lam_ref is not None:
                ax.plot(lam_ref[i], lam_ref[j], 'wo', markersize=10)
            if lambda_label is None:
                label1 = r'$\lambda_{' + str(i + 1) + '}$'
                label2 = r'$\lambda_{' + str(j + 1) + '}$'
            else:
                label1 = lambda_label[i]
                label2 = lambda_label[j]
            ax.set_xlabel(label1, fontsize=20)
            ax.set_ylabel(label2, fontsize=20)
            ax.tick_params(axis='both', which='major', labelsize=14)
            label_cbar = r'$\rho_{\lambda_{' + str(i + 1) + '}, '
            label_cbar += r'\lambda_{' + str(j + 1) + '}' + '}$ (Lebesgue)'
            cb = fig.colorbar(quadmesh, ax=ax, label=label_cbar)
            cb.ax.tick_params(labelsize=14)
            cb.set_label(label_cbar, size=20)
            plt.axis([lam_domain[i][0], lam_domain[i][1], lam_domain[j][0],
                      lam_domain[j][1]])
            plt.tight_layout()
            fig.savefig(filename + "_2D_" + str(i) + "_" + str(j) +
                        file_extension, transparent=True)
            if interactive:
                plt.show()
            else:
                plt.close()

        if plot_surface:
            for k, (i, j) in enumerate(pairs):
                fig = plt.figure(k)
                ax = fig.gca(projection='3d')
                X = bins[i][:-1] + np.diff(bins[i]) / 2
                Y = bins[j][:-1] + np.diff(bins[j]) / 2
                X, Y = np.meshgrid(X, Y, indexing='ij')
                surf = ax.plot_surface(X, Y, marginals[(i, j)], rstride=1,
                                       cstride=1, cmap=cm.coolwarm, linewidth=0,
                                       antialiased=False)
                ax.zaxis.set_major_locator(LinearLocator(10))
                ax.zaxis.set_major_formatter(FormatStrFormatter('%.02f'))
                ax.set_xlabel(r'$\lambda_{' + str(i + 1) + '}$')
                ax.set_ylabel(r'$\lambda_{' + str(j + 1) + '}$')
                ax.set_zlabel(r'$P$')
                plt.backgroundcolor = 'w'
                fig.colorbar(surf, shrink=0.5, aspect=5, label=r'$P$')
                plt.tight_layout()
                fig.savefig(filename + "_surf_" + str(i) + "_" + str(j) +
                            file_extension, transparent=True)

                if interactive:
                    plt.show()
                else:
                    plt.close()
                plt.clf()
    comm.barrier()


def smooth_marginals_1D(marginals, bins, sigma=10.0):
    """
    This function smooths 1D marginal probabilities calculated from histograms.

    :param marginals: 1D marginal probabilities
    :type marginals: dictionary with int as keys and :class:`~numpy.ndarray` of
        shape (nbins+1,) as values
    :param bins: Endpoints of bins used in calculating marginals
    :type bins: :class:`~numpy.ndarray` of shape (nbins+1,)
    :param sigma: Smoothing parameter in each direction.
    :type sigma: float or :class:`~numpy.ndarray` of shape (ndim,)
    :rtype: dict
    :returns: marginals_smooth
    """
    from scipy.fftpack import fftshift, ifft, fft

    if isinstance(sigma, float) or isinstance(sigma, int):
        sigma = sigma * np.ones(len(bins), dtype=np.int)
    marginals_smooth = {}
    index = sorted(copy.deepcopy(list(marginals.keys())))
    for i in index:
        nx = len(bins[i]) - 1
        dx = bins[i][1] - bins[i][0]
        augx = int(math.ceil(3 * sigma[i] / dx))
        x_kernel = np.linspace(-nx * dx / 2, nx * dx / 2, nx)
        kernel = np.exp(-(x_kernel / sigma[i])**2)
        aug_kernel = np.zeros((nx + 2 * augx,))
        aug_marginals = np.zeros((nx + 2 * augx,))

        aug_kernel[augx:augx + nx] = kernel
        aug_marginals[augx:augx + nx] = marginals[i]

        aug_kernel = fftshift(aug_kernel)

        aug_marginals_smooth = np.real(
            ifft(fft(aug_kernel) * fft(aug_marginals)))
        marginals_smooth[i] = aug_marginals_smooth[augx:augx + nx]
        marginals_smooth[i] = marginals_smooth[i] / np.sum(marginals_smooth[i])

    return marginals_smooth


def smooth_marginals_2D(marginals, bins, sigma=10.0):
    """
    This function smooths 2D marginal probabilities calculated from histograms.

    :param marginals: 2D marginal probabilities
    :type marginals: dictionary with tuples of 2 integers as keys and
        :class:`~numpy.ndarray` of shape (nbins+1,) as values
    :param bins: Endpoints of bins used in calculating marginals
    :type bins: :class:`~numpy.ndarray` of shape (nbins+1,)
    :param sigma: Smoothing parameter in each direction.
    :type sigma: float or :class:`~numpy.ndarray` of shape (ndim,)
    :rtype: dict
    :returns: marginals_smooth
    """
    from scipy.fftpack import fftshift, ifft2, fft2

    if isinstance(sigma, float) or isinstance(sigma, int):
        sigma = sigma * np.ones(len(bins), dtype=np.int)
    marginals_smooth = {}
    pairs = sorted(copy.deepcopy(list(marginals.keys())))
    for (i, j) in pairs:
        nx = len(bins[i]) - 1
        ny = len(bins[j]) - 1
        dx = bins[i][1] - bins[i][0]
        dy = bins[j][1] - bins[j][0]

        augx = int(math.ceil(3 * sigma[i] / dx))
        augy = int(math.ceil(3 * sigma[j] / dy))

        x_kernel = np.linspace(-nx * dx / 2, nx * dx / 2, nx)
        y_kernel = np.linspace(-ny * dy / 2, ny * dy / 2, ny)
        X, Y = np.meshgrid(x_kernel, y_kernel, indexing='ij')

        kernel = np.exp(-(X / sigma[i])**2 - (Y / sigma[j])**2)
        aug_kernel = np.zeros((nx + 2 * augx, ny + 2 * augy))
        aug_marginals = np.zeros((nx + 2 * augx, ny + 2 * augy))

        aug_kernel[augx:augx + nx, augy:augy + ny] = kernel
        aug_marginals[augx:augx + nx, augy:augy + ny] = marginals[(i, j)]

        aug_kernel = fftshift(aug_kernel, 0)
        aug_kernel = fftshift(aug_kernel, 1)
        aug_marginals_smooth = ifft2(fft2(aug_kernel) * fft2(aug_marginals))
        aug_marginals_smooth = np.real(aug_marginals_smooth)
        marginals_smooth[(i, j)] = aug_marginals_smooth[augx:augx + nx,
                                                        augy:augy + ny]
        marginals_smooth[(i, j)] = marginals_smooth[(i,
                                                     j)] / np.sum(marginals_smooth[(i, j)])

    return marginals_smooth


def plot_2D_marginal_contours(marginals, bins, sample_set,
                              contour_num=8,
                              lam_ref=None, lam_refs=None,
                              plot_domain=None,
                              interactive=False,
                              lambda_label=None,
                              contour_font_size=20,
                              filename="file",
                              file_extension=".png"):
    """
    This makes contour plots of every pair of marginals (or joint in 2d case)
    of input probability measure on a rectangular grid.
    If the sample_set object is a discretization object, we assume
    that the probabilities to be plotted are from the input space.

    .. note::

        Do not specify the file extension in the file name.

    :param marginals: 2D marginal probabilities
    :type marginals: dictionary with tuples of 2 integers as keys and
        :class:`~numpy.ndarray` of shape (nbins+1,) as values
    :param bins: Endpoints of bins used in calculating marginals
    :type bins: :class:`~numpy.ndarray` of shape (nbins+1,2)
    :param sample_set: Object containing samples and probabilities
    :type sample_set: :class:`~bet.sample.sample_set_base`
        or :class:`~bet.sample.discretization`
    :param filename: Prefix for output files.
    :type filename: str
    :param lam_ref: True parameters.
    :type lam_ref: :class:`~numpy.ndarray` of shape (ndim,) or None
    :param interactive: Whether or not to display interactive plots.
    :type interactive: bool
    :param lambda_label: Label for each parameter for plots.
    :type lambda_label: list of length nbins of strings or None
    :param string file_extension: file extenstion

    """
    if isinstance(sample_set, sample.discretization):
        sample_obj = sample_set._input_sample_set
    elif isinstance(sample_set, sample.sample_set_base):
        sample_obj = sample_set
    else:
        raise bad_object("Improper sample object")

    if lam_ref is None:
        lam_ref = sample_obj._reference_value

    lam_domain = sample_obj.get_domain()

    matplotlib.rcParams['xtick.direction'] = 'out'
    matplotlib.rcParams['ytick.direction'] = 'out'
    matplotlib.rcParams.update({'figure.autolayout': True})

    if comm.rank == 0:
        pairs = sorted(copy.deepcopy(list(marginals.keys())))
        for k, (i, j) in enumerate(pairs):
            fig = plt.figure(k)
            ax = fig.add_subplot(111)
            boxSize = (bins[i][1] - bins[i][0]) * (bins[j][1] - bins[j][0])
            nx = len(bins[i]) - 1
            ny = len(bins[j]) - 1
            dx = bins[i][1] - bins[i][0]
            dy = bins[j][1] - bins[j][0]

            x_kernel = np.linspace(-nx * dx / 2, nx * dx / 2, nx)
            y_kernel = np.linspace(-ny * dy / 2, ny * dy / 2, ny)
            X, Y = np.meshgrid(x_kernel, y_kernel, indexing='ij')
            quadmesh = ax.contour(marginals[(i, j)].transpose() / boxSize,
                                  contour_num, colors='k',
                                  extent=[lam_domain[i][0], lam_domain[i][1],
                                          lam_domain[j][0], lam_domain[j][1]], origin='lower',
                                  vmax=marginals[(i, j)].max() / boxSize, vmin=0,
                                  aspect='auto')
            if lam_refs is not None:
                ax.plot(lam_refs[:, i], lam_refs[:, j], 'wo', markersize=20)
            if lam_ref is not None:
                ax.plot(lam_ref[i], lam_ref[j], 'ko', markersize=20)
            if lambda_label is None:
                label1 = r'$\lambda_{' + str(i + 1) + '}$'
                label2 = r'$\lambda_{' + str(j + 1) + '}$'
            else:
                label1 = lambda_label[i]
                label2 = lambda_label[j]
            ax.set_xlabel(label1, fontsize=30)
            ax.set_ylabel(label2, fontsize=30)
            ax.tick_params(axis='both', which='major',
                           labelsize=20)
            plt.clabel(quadmesh, fontsize=contour_font_size,
                       inline=1)

            if plot_domain is None:
                plt.axis([lam_domain[i][0], lam_domain[i][1],
                          lam_domain[j][0], lam_domain[j][1]])
            else:
                plt.axis([plot_domain[i][0], plot_domain[i][1],
                          plot_domain[j][0], plot_domain[j][1]])
            plt.tight_layout()
            fig.savefig(filename + "_2D_contours_" + str(i) + "_" + str(j) +
                        file_extension, transparent=True)
            if interactive:
                plt.show()
            else:
                plt.close()

    comm.barrier()


def plot_1d_marginal_densities(sets, i, interval=None, num_points=1000, label=None, sets_label=None,
                               sets_label_initial=None, title=None, initials=True, inputs=True,
                               interactive=True, savefile=None):
    """
    Plot 1D marginal probability density functions in direction `i`. Useful for visualizing
    solutions of density-based inverse problems.

    :param sets: Object containing sample sets to plot marginals for.
    :type sets: :class:`bet.sample.sample_set` or :class:`bet.sample.discretization` or list or tuple of these
    :param i: index of direction to take marginal
    :type i: int
    :param interval: Interval over which to plot.
    :type interval: list
    :param num_points: Number of points to evaluate PDFs at.
    :type num_points: int
    :param label: Label for parameter i
    :type label: str
    :param sets_label: Labels for sets
    :type sets_label: List or tuple of strings.
    :param sets_label_initial: Labels for sets' initial probabilities
    :type sets_label_initial: List or tuple of strings.
    :param title: "Title for plot"
    :type title: str
    :param initials: Whether or not to plot initial probabilities
    :type initials: bool
    :param inputs: Whether to use input or output sample sets for disretizations
    :type inputs: bool
    :param interactive: Whether or not to show interactive figure
    :type interactive: bool
    :param savefile: filename to save to
    :type savefile: str
    """
    if isinstance(sets, sample.sample_set) or isinstance(sets, sample.discretization):
        sets = [sets]
    new_sets = []
    for s in sets:
        if isinstance(s, sample.sample_set):
            new_sets.append(s)
        elif isinstance(s, sample.discretization):
            if inputs:
                new_sets.append(s.get_input_sample_set())
            else:
                new_sets.append(s.get_output_sample_set())
        else:
            raise bad_object("One of the input sets does not contain a sample set.")
    sets = new_sets

    # set labels
    if label is None and sets[0].get_labels() is not None:
        label = sets[0].get_labels()[i]
    elif label is None:
        label = 'Parameter ' + str(i)

    if sets_label is None:
        sets_label = []
        for j, s in enumerate(sets):
            if s.get_labels() is None:
                sets_label.append('Set ' + str(j) + ' Updated')
            else:
                sets_label.append(s.get_labels()[i])
    if sets_label_initial is None:
        sets_label_initial = []
        for j, s in enumerate(sets):
            if s.get_labels() is None:
                sets_label_initial.append('Set ' + str(j) + ' Initial')
            else:
                sets_label_initial.append(s.get_labels()[i] + ' Initial')

    if interval is None:
        x_min = np.inf
        x_max = -np.inf
        for s in sets:
            min1 = np.min(s.get_values()[:, i])
            max1 = np.max(s.get_values()[:, i])
            if min1 < x_min:
                x_min = min1
            if max1 > x_max:
                x_max = max1
        delt = 0.25 * (x_max - x_min)
        x = np.linspace(x_min - delt, x_max + delt, 100)
    else:
        x = np.linspace(interval[0], interval[1], num_points)

    # plot marginals
    plt.rcParams.update({'font.size': 22})
    plt.rcParams.update({'axes.linewidth': 2})
    fig = plt.figure(figsize=(10, 10))
    for k, s in enumerate(sets):
        if s.get_prob_type_init() is not None and initials:
            mar = s.marginal_pdf_init(x, i)
            plt.plot(x, mar, label=sets_label_initial[k], linewidth=4)
        if s.get_prob_type() is not None:
            mar = s.marginal_pdf(x, i)
            plt.plot(x, mar, label=sets_label[k], linewidth=4, linestyle='dashed')
    plt.xlabel(label)
    plt.ylabel("PDF")
    if type(title) is str:
        plt.title(title)
    plt.legend(fontsize=16)
    if interactive:
        plt.show()
    if savefile is not None:
        plt.savefig(savefile)



# Copyright (C) 2014-2020 The BET Development Team

# -*- coding: utf-8 -*-
import numpy as np
from dolfin import *
from meshDS import meshDS
from projectKL import projectKL
from poissonRandField import solvePoissonRandomField
import scipy.io as sio
import sys
import ast


def my_model(parameter_sample):
    KL_mdat = sio.loadmat("KL_expansion")
    KL_eigen_funcs = KL_mdat['KL_eigen_funcs']
    KL_eigen_vals = KL_mdat['KL_eigen_vals']

    param_sample_num = parameter_sample[0]
    # the samples are the coefficients of the KL expansion typically denoted
    # by xi_k
    xi_k = np.array(parameter_sample[1:])
    xi_k = xi_k.astype(float)

    # number of KL expansion terms.
    numKL = np.size(parameter_sample[1:])
    print('hi')
    print(xi_k)

    '''
    ++++++++++++++++ Steps in Computing the Numerical KL Expansion ++++++++++
    We proceed by loading the mesh and defining the function space for which
    the eigenfunctions are defined upon.

    Then, we define the covariance kernel which requires correlation lengths
    and a standard deviation.

    We then compute the truncated KL expansion.
    +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    '''
    # Step 1: Set up the Mesh and Function Space

    mesh = Mesh("Lshaped.xml")
    #mesh = RectangleMesh(0,0,10,10,20,20)

    # Plot the mesh for visual check
    # plot(mesh,interactive=True)

    # initialize the mesh to generate connectivity
    mesh.init()

    # Random field is projected on the space of Hat functions in the mesh
    V = FunctionSpace(mesh, "Lagrange", 1)

    '''
    ++++++++++++++++ Steps in Solving Poisson with the KL fields ++++++++++++
    First set up the necessary variables and boundary conditions for the
    problem (done in Compute_Save_KL in the BET script).

    Then create the QoI maps defined by average values over some part of the
    physical domain (done here).

    Loop through the sample fields and create the permeability defined by the
    exponential of the KL field (i.e., the KL expansion represents the log of
    the permeability) (done by launcher).

    For each sample field, call the PoissonRandomField solver. (done here)
    +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    '''
    # permeability
    perm_k = Function(V)

    # Create Boundary Conditions -- Dirichlet on left and bottom boundary.
    # Left Dirichlet Bc
    def left_boundary(x, on_boundary):
        """TODO: Docstring for left_boundary.

        :x: TODO
        :on_boundary: TODO
        :returns: TODO

        """
        tol = 1e-14
        return on_boundary and abs(x[0]) < tol
    Gamma_0 = DirichletBC(V, Constant(0.0), left_boundary)

    def bottom_boundary(x, on_boundary):
        """TODO: Docstring for left_boundary.

        :x: TODO
        :on_boundary: TODO
        :returns: TODO

        """
        tol = 1e-14
        return on_boundary and abs(x[1]) < tol
    Gamma_1 = DirichletBC(V, Constant(0.0), bottom_boundary)
    bcs = [Gamma_0, Gamma_1]

    # Setup the QoI class
    class CharFunc(Expression):
        def __init__(self, region):
            self.a = region[0]
            self.b = region[1]
            self.c = region[2]
            self.d = region[3]

        def eval(self, v, x):
            v[0] = 0
            if (x[0] >= self.a) & (x[0] <= self.b) & (
                    x[1] >= self.c) & (x[1] <= self.d):
                v[0] = 1
            return v

    # Define the QoI maps
    Chi_1 = CharFunc([0.75, 1.25, 7.75, 8.25])
    Chi_2 = CharFunc([7.75, 8.25, 0.75, 1.25])

    QoI_samples = np.zeros([1, 2])

    # For each sample solve the PDE
    f = Constant(-1.0)  # forcing of Poisson

    # create a temp array to store logPerm as sum of KL expansions
    # logPerm is log permeability
    logPerm = np.zeros((mesh.num_vertices()), dtype=float)
    for kl in range(0, numKL):
        logPerm += xi_k[kl] * \
            np.sqrt(KL_eigen_vals[0, kl]) * KL_eigen_funcs[kl, :]

    # permiability is the exponential of log permeability logPerm
    perm_k_array = 0.1 + np.exp(logPerm)

    # use dof_to_vertex map to map values to the function space
    perm_k.vector()[:] = perm_k_array

    # solve Poisson with this random field using FEM
    u = solvePoissonRandomField(perm_k, mesh, 1, f, bcs)

    # Compute QoI
    QoI_samples[0, 0] = assemble(u * Chi_1 * dx)
    QoI_samples[0, 1] = assemble(u * Chi_2 * dx)

    io_file_name = 'QoI_sample' + param_sample_num
    io_mdat = dict()
    io_mdat['output'] = QoI_samples
    sio.savemat(io_file_name, io_mdat)


def usage():
    print("usage: [io_file]")


if __name__ == "__main__":
    if len(sys.argv) > 1:
        my_model(sys.argv[1:])
    else:
        usage()



# Copyright (C) 2014-2020 The BET Development Team

"""
This module contains tests for :module:`bet.postProcess.plotVoronoi`.


Tests for Voronoi plotting.
"""

import unittest
import bet.calculateP.calculateP as calcP
import bet.calculateP.simpleFunP as simpleFunP
import bet.postProcess.plotP as plotP
import bet.postProcess.plotVoronoi as plotVoronoi
import numpy as np
import scipy.spatial as spatial
import numpy.testing as nptest
import bet.util as util
from bet.Comm import comm
import os
import bet.sample as sample


@unittest.skipIf(comm.size > 1, 'Only run in serial')
class Test_plot_1D_voronoi(unittest.TestCase):
    """
    Test :meth:`bet.postProcess.plotVoronoi.plot_1D_voronoi`
    for a 1D parameter space.
    """

    def setUp(self):
        """
        Set up problem.
        """
        emulated_input_samples = sample.sample_set(1)
        emulated_input_samples.set_domain(np.array([[0.0, 1.0]]))

        num_samples = 100

        emulated_input_samples.set_values_local(np.linspace(emulated_input_samples.get_domain()[0][0],
                                                            emulated_input_samples.get_domain()[
            0][1],
            num_samples + 1))

        emulated_input_samples.set_probabilities_local(1.0 / float(comm.size) * (1.0 / float(
            emulated_input_samples.get_values_local().shape[0]))
            * np.ones((emulated_input_samples.get_values_local().shape[0],)))
        emulated_input_samples.set_volumes_local(1.0 / float(comm.size) * (1.0 / float(
            emulated_input_samples.get_values_local().shape[0]))
            * np.ones((emulated_input_samples.get_values_local().shape[0],)))
        emulated_input_samples.check_num()

        self.samples = emulated_input_samples

    def test_plot(self):
        """
        Test :meth:`bet.postProcess.plotVoronoi.plot_1D_voronoi`
        for a 1D parameter space.
        """
        try:
            plotVoronoi.plot_1D_voronoi(self.samples,
                                        filename="file", interactive=False)
            go = True
            if os.path.exists("file.png"):
                os.remove("file.png")
        except (RuntimeError, TypeError, NameError):
            go = False
        nptest.assert_equal(go, True)


@unittest.skipIf(comm.size > 1, 'Only run in serial')
class Test_plot_2D_voronoi(unittest.TestCase):
    """
    Test :meth:`bet.postProcess.plotP.plot_2D_voronoi`
    for a 2D parameter space.
    """

    def setUp(self):
        """
        Set up problem.
        """
        emulated_input_samples = sample.sample_set(2)
        emulated_input_samples.set_domain(np.array([[0.0, 1.0], [0.0, 1.0]]))

        emulated_input_samples.set_values_local(util.meshgrid_ndim((np.linspace(emulated_input_samples.get_domain()[0][0],
                                                                                emulated_input_samples.get_domain()[0][1], 10),
                                                                    np.linspace(emulated_input_samples.get_domain()[1][0],
                                                                                emulated_input_samples.get_domain()[1][1], 10))))

        emulated_input_samples.set_probabilities_local(1.0 / float(comm.size) *
                                                       (1.0 / float(emulated_input_samples.get_values_local().shape[0])) *
                                                       np.ones((emulated_input_samples.get_values_local().shape[0],)))
        emulated_input_samples.set_volumes_local(1.0 / float(comm.size) *
                                                 (1.0 / float(emulated_input_samples.get_values_local().shape[0])) *
                                                 np.ones((emulated_input_samples.get_values_local().shape[0],)))
        emulated_input_samples.check_num()

        self.samples = emulated_input_samples

    def test_plot(self):
        """
        Test :meth:`bet.postProcess.plotP.plot_2D_voronoi`
        for a 2D parameter space.
        """
        try:
            plotVoronoi.plot_2D_voronoi(self.samples,
                                        filename="file", interactive=False)
            go = True
            if os.path.exists("file.png"):
                os.remove("file.png")
        except (RuntimeError, TypeError, NameError):
            go = False
        nptest.assert_equal(go, True)



# Copyright (C) 2014-2020 The BET Development Team

"""
This module contains unittests for :mod:`~bet.sampling.useLUQ`
"""

import unittest
import os
import pyDOE
import numpy.testing as nptest
import numpy as np
import scipy.io as sio
import bet
import bet.sampling.basicSampling as bsam
import bet.sampling.useLUQ as useLUQ
from bet.Comm import comm
import bet.sample
from bet.sample import sample_set
from bet.sample import discretization as disc
import collections

try:
    from luq.luq import LUQ
    has_luq = True
except ImportError:
    has_luq = False


@unittest.skipIf(not has_luq, 'LUQ is not installed.')
@unittest.skipIf(comm.size > 1, 'Only run in serial')
class Test_useLUQ(unittest.TestCase):
    """
    Testing ``bet.sampling.useLUQ.useLUQ``, interfacing with a model.
    """
    def setUp(self):
        np.random.seed(123456)
        self.p_set = bsam.random_sample_set(rv=[['uniform', {'loc': .01, 'scale': 0.114}],
                                        ['uniform', {'loc': .05, 'scale': 1.45}]],
                                       input_obj=2, num_samples=20)

        self.o_set = bsam.random_sample_set(rv=[['beta', {'a': 2, 'b': 2, 'loc': .01, 'scale': 0.114}],
                                           ['beta', {'a': 2, 'b': 2, 'loc': .05, 'scale': 1.45}]],
                                       input_obj=2, num_samples=20)
        time_start = 2.0  # 0.5
        time_end = 6.5  # 40.0
        num_time_preds = int((time_end - time_start) * 100)
        self.times = np.linspace(time_start, time_end, num_time_preds)

        self.luq = useLUQ.useLUQ(predict_set=self.p_set, obs_set=self.o_set, lb_model=useLUQ.myModel, times=self.times)
        self.luq.setup()

        time_start_idx = 0
        time_end_idx = len(self.luq.times) - 1
        self.luq.clean_data(time_start_idx=time_start_idx, time_end_idx=time_end_idx,
                            num_clean_obs=20, tol=5.0e-2, min_knots=3, max_knots=12)
        self.luq.dynamics(cluster_method='kmeans', kwargs={'n_clusters': 3, 'n_init': 10})
        self.luq.learn_qois_and_transform(num_qoi=2)
        self.disc1, self.disc2 = self.luq.make_disc()

    def test_nums(self):
        """
        Check the number of samples.
        """
        self.disc1.check_nums()
        self.disc2.check_nums()

    def test_dims(self):
        """
        Check the dimensions.
        """
        assert self.disc1.get_output_sample_set().get_dim() == 2
        assert self.disc2.get_output_sample_set().get_dim() == 2

    def test_sets(self):
        """
        Check the sets
        """
        assert self.disc1.get_input_sample_set() == self.p_set
        assert self.disc2.get_input_sample_set() == self.o_set
        assert self.disc1.get_output_observed_set() == self.disc2.get_output_sample_set()

    def test_saving(self):
        """
        Test saving.
        """
        savefile = 'test_save_useLUQ'
        self.luq.save(savefile)
        loaded = bet.util.load_object(file_name=savefile)
        disc1, disc2 = loaded.make_disc()
        assert disc1 == self.disc1
        assert disc2 == self.disc2

        if comm.rank == 0:
            os.remove(savefile + '.p')



# Copyright (C) 2014-2020 The BET Development Team

# -*- coding: utf-8 -*-
import numpy as np

# Define a model that is a linear QoI map


def my_model(parameter_samples):
    Q_map = np.array([[0.506, 0.463], [0.253, 0.918], [0.085, 0.496]])
    #Q_map = np.array([[0.506], [0.253], [0.085]])
    QoI_samples = np.dot(parameter_samples, Q_map)
    return QoI_samples



# Copyright (C) 2014-2020 The BET Development Team

"""
This module contains unittests for :mod:`~bet.Comm`
"""

import unittest
import bet.Comm as Comm
from pkgutil import iter_modules


class Test_comm_for_no_mpi4py(unittest.TestCase):
    """
    Test :class:`bet.Comm.comm_for_no_mpi4py`.
    """

    def setUp(self):
        self.comm = Comm.comm_for_no_mpi4py()

    def test_Get_size(self):
        self.assertEqual(self.comm.Get_size(), 1)

    def test_Get_rank(self):
        self.assertEqual(self.comm.Get_rank(), 0)

    def test_allgather(self):
        thing = list(range(4))
        self.assertEqual(self.comm.allgather(thing), [thing])

    def test_allreduce(self):
        thing = 4
        self.assertEqual(self.comm.allreduce(thing, op=None), thing)

    def test_bcast(self):
        thing = list(range(4))
        self.assertEqual(self.comm.bcast(thing, root=0), thing)

    def test_Allgather(self):
        thing = list(range(4))
        self.assertEqual(self.comm.Allgather(thing), thing)

    def test_Allreduce(self):
        thing1 = list(range(4))
        thing2 = list(range(4))
        self.assertEqual(self.comm.Allreduce(thing1, thing2,
                                             op=None), thing1)

    def test_Bcast(self):
        thing = list(range(4))
        self.assertEqual(self.comm.Bcast(thing, root=0), thing)

    def test_Scatter(self):
        thing1 = list(range(4))
        thing2 = list(range(4))
        self.assertEqual(self.comm.Scatter(thing1, thing2,
                                           root=0), thing1)


class Test_Comm(unittest.TestCase):
    """
    Test :mod:`bet.Comm`
    """

    def test(self):
        if 'mpi4py' in (name for loader, name, ispkg in iter_modules()):
            pass
        else:
            self.assertEqual(Comm.comm.size, 1)
            self.assertEqual(Comm.comm.rank, 0)


class Test_MPI_for_no_mpi4py(unittest.TestCase):
    """
    Test :class:`bet.Comm.MPI_fort_no_mpi4py`
    """

    def test(self):
        MPI_no = Comm.MPI_for_no_mpi4py()
        self.assertEqual(MPI_no.SUM, None)
        self.assertEqual(MPI_no.DOUBLE, float)
        self.assertEqual(MPI_no.INT, int)



# Copyright (C) 2014-2020 The BET Development Team

"""
This module contains tests for :module:`bet.postProcess.plotP`.


Tests for correct computation of marginals and plotting.
"""

import unittest
import bet.calculateP.calculateP as calcP
import bet.calculateP.simpleFunP as simpleFunP
import bet.postProcess.plotP as plotP
import numpy as np
import scipy.spatial as spatial
import numpy.testing as nptest
import bet.util as util
from bet.Comm import comm
import os
import bet.sample as sample
import bet.calculateP.calculateR as calculateR
import bet.sampling.basicSampling as bsam


class Test_calc_marg_1D(unittest.TestCase):
    """
    Test :meth:`bet.postProcess.plotP.calculate_1D_marginal_probs`
    for a 1D parameter space.
    """

    def setUp(self):
        """
        Set up problem.
        """
        emulated_input_samples = sample.sample_set(1)
        emulated_input_samples.set_domain(np.array([[0.0, 1.0]]))

        num_samples = 1000

        emulated_input_samples.set_values_local(
            np.linspace(emulated_input_samples.get_domain()[0][0],
                        emulated_input_samples.get_domain()[0][1],
                        num_samples + 1))

        emulated_input_samples.set_probabilities_local(
            1.0 / float(comm.size) * (1.0 / float(
                emulated_input_samples.get_values_local().shape[0])) *
            np.ones((emulated_input_samples.get_values_local().shape[0],)))

        emulated_input_samples.check_num()

        self.samples = emulated_input_samples

    def test_1_bin(self):
        """
        Test that marginals sum to 1 and have correct shape.
        """
        (bins, marginals) = plotP.calculate_1D_marginal_probs(self.samples,
                                                              nbins=1)

        nptest.assert_almost_equal(marginals[0][0], 1.0)
        nptest.assert_equal(marginals[0].shape, (1,))

    def test_10_bins(self):
        """
        Test that marginals sum to 1 and have correct shape.
        """
        (bins, marginals) = plotP.calculate_1D_marginal_probs(self.samples,
                                                              nbins=10)

        nptest.assert_almost_equal(np.sum(marginals[0]), 1.0)
        nptest.assert_equal(marginals[0].shape, (10,))


class Test_calc_marg_2D(unittest.TestCase):
    """
    Test :meth:`bet.postProcess.plotP.calculate_1D_marginal_probs` and
    :meth:`bet.postProcess.plotP.calculate_2D_marginal_probs` for a 2D
    parameter space.
    """

    def setUp(self):
        """
        Set up problem.
        """
        emulated_input_samples = sample.sample_set(2)
        emulated_input_samples.set_domain(np.array([[0.0, 1.0], [0.0, 1.0]]))

        emulated_input_samples.set_values_local(
            util.meshgrid_ndim((np.linspace(
                emulated_input_samples.get_domain()[0][0],
                emulated_input_samples.get_domain()[0][1],
                10),
                np.linspace(
                emulated_input_samples.get_domain()[1][0],
                emulated_input_samples.get_domain()[1][1],
                10)
            ))
        )

        emulated_input_samples.set_probabilities_local(1.0 / float(comm.size) *
                                                       (
            1.0 / float(
                emulated_input_samples.get_values_local().shape[0])
        ) *
            np.ones(
            (emulated_input_samples.get_values_local().shape[0],)
        )
        )
        emulated_input_samples.check_num()

        self.samples = emulated_input_samples

    def test_1_bin_1D(self):
        """
        Test that 1D marginals sum to 1 and have right shape.
        """
        (bins, marginals) = plotP.calculate_1D_marginal_probs(self.samples,
                                                              nbins=1)

        nptest.assert_almost_equal(marginals[0][0], 1.0)
        nptest.assert_almost_equal(marginals[1][0], 1.0)
        nptest.assert_equal(marginals[0].shape, (1,))
        nptest.assert_equal(marginals[1].shape, (1,))

    def test_10_bins_1D(self):
        """
        Test that 1D marginals sum to 1 and have right shape.
        """
        (bins, marginals) = plotP.calculate_1D_marginal_probs(self.samples,
                                                              nbins=10)

        nptest.assert_almost_equal(np.sum(marginals[0]), 1.0)
        nptest.assert_almost_equal(np.sum(marginals[1]), 1.0)
        nptest.assert_equal(marginals[0].shape, (10,))

    def test_1_bin_2D(self):
        """
        Test that 2D marginals sum to 1 and have right shape.
        """
        (bins, marginals) = plotP.calculate_2D_marginal_probs(self.samples,
                                                              nbins=1)

        nptest.assert_almost_equal(marginals[(0, 1)][0], 1.0)
        nptest.assert_equal(marginals[(0, 1)].shape, (1, 1))

    def test_10_bins_2D(self):
        """
        Test that 2D marginals sum to 1 and have right shape.
        """
        (bins, marginals) = plotP.calculate_2D_marginal_probs(self.samples,
                                                              nbins=10)

        nptest.assert_almost_equal(np.sum(marginals[(0, 1)]), 1.0)
        nptest.assert_equal(marginals[(0, 1)].shape, (10, 10))

    def test_5_10_bins_2D(self):
        """
        Test that 1D marginals sum to 1 and have right shape.
        """
        (bins, marginals) = plotP.calculate_2D_marginal_probs(self.samples,
                                                              nbins=[5, 10])

        nptest.assert_almost_equal(np.sum(marginals[(0, 1)]), 1.0)
        nptest.assert_equal(marginals[(0, 1)].shape, (5, 10))

    def test_1D_smoothing(self):
        """
        Test :meth:`bet.postProcess.plotP.smooth_marginals_1D`.
        """
        (bins, marginals) = plotP.calculate_1D_marginal_probs(self.samples,
                                                              nbins=10)

        marginals_smooth = plotP.smooth_marginals_1D(marginals, bins,
                                                     sigma=10.0)

        nptest.assert_equal(marginals_smooth[0].shape, marginals[0].shape)
        nptest.assert_almost_equal(np.sum(marginals_smooth[0]), 1.0)

    def test_2D_smoothing(self):
        """
        Test :meth:`bet.postProcess.plotP.smooth_marginals_2D`.
        """
        (bins, marginals) = plotP.calculate_2D_marginal_probs(self.samples,
                                                              nbins=10)

        marginals_smooth = plotP.smooth_marginals_2D(marginals, bins,
                                                     sigma=10.0)

        nptest.assert_equal(marginals_smooth[(0, 1)].shape,
                            marginals[(0, 1)].shape)
        nptest.assert_almost_equal(np.sum(marginals_smooth[(0, 1)]), 1.0)

    def test_plot_marginals_1D(self):
        """
        Test :meth:`bet.postProcess.plotP.plot_1D_marginal_probs`.
        """
        (bins, marginals) = plotP.calculate_1D_marginal_probs(self.samples,
                                                              nbins=10)

        try:
            plotP.plot_1D_marginal_probs(marginals, bins, self.samples,
                                         filename="file", interactive=False)
            go = True
            if os.path.exists("file_1D_0.png") and comm.rank == 0:
                os.remove("file_1D_0.png")
            if os.path.exists("file_1D_1.png") and comm.rank == 0:
                os.remove("file_1D_1.png")
        except (RuntimeError, TypeError, NameError):
            go = False
        nptest.assert_equal(go, True)

    def test_plot_marginals_2D(self):
        """
        Test :meth:`bet.postProcess.plotP.plot_2D_marginal_probs`.
        """
        (bins, marginals) = plotP.calculate_2D_marginal_probs(self.samples,
                                                              nbins=10)
        marginals[(0, 1)][0][0] = 0.0
        marginals[(0, 1)][0][1] *= 2.0
        try:
            plotP.plot_2D_marginal_probs(marginals, bins, self.samples,
                                         filename="file", interactive=False)
            plotP.plot_2D_marginal_probs(marginals, bins, self.samples, plot_surface=True,
                                         filename="file", interactive=False)
            go = True
            if os.path.exists("file_2D_0_1.png") and comm.rank == 0:
                os.remove("file_2D_0_1.png")
            if os.path.exists("file_surf_0_1.png") and comm.rank == 0:
                os.remove("file_surf_0_1.png")
        except (RuntimeError, TypeError, NameError):
            go = False
        nptest.assert_equal(go, True)

    def test_plot_2D_marginal_contours(self):
        """
        Test :meth:`bet.postProcess.plotP.plot_2D_marginal_contours`.
        """
        (bins, marginals) = plotP.calculate_2D_marginal_probs(self.samples,
                                                              nbins=10)
        marginals[(0, 1)][0][0] = 0.0
        marginals[(0, 1)][0][1] *= 2.0
        try:
            plotP.plot_2D_marginal_probs(marginals, bins, self.samples,
                                         filename="file", interactive=False)
            go = True
            if os.path.exists("file_2D_contours_0_1.png") and comm.rank == 0:
                os.remove("file_2D_contours_0_1.png")
        except (RuntimeError, TypeError, NameError):
            go = False
        nptest.assert_equal(go, True)


@unittest.skipIf(comm.size > 1, 'Only run in serial')
class Test_plot_1d_marginal_densities(unittest.TestCase):
    """
    Test :meth:`bet.postProcess.plotP.plot_1d_marginal_densities`.
    """
    def setUp(self):
        def my_model(parameter_samples):
            Q_map = np.array([[0.506, 0.463], [0.253, 0.918], [0.685, 0.496]])
            QoI_samples = np.dot(parameter_samples, Q_map)
            return QoI_samples

        sampler = bsam.sampler(my_model)
        sampler.random_sample_set(rv=[['norm', {'loc': 2, 'scale': 3}],
                                      ['uniform', {'loc': 2, 'scale': 3}],
                                      ['beta', {'a': 2, 'b': 2}]], input_obj=3, num_samples=1000)
        sampler.compute_qoi_and_create_discretization()

        sampler2 = bsam.sampler(my_model)
        sampler2.random_sample_set(rv=[['norm', {'loc': 1, 'scale': 2}],
                                       ['uniform', {'loc': 2, 'scale': 2}],
                                       ['beta', {'a': 2, 'b': 3}]], input_obj=3, num_samples=1000)
        sampler2.compute_qoi_and_create_discretization()

        sampler.discretization.set_output_observed_set(sampler2.discretization.get_output_sample_set())
        self.disc1 = sampler.discretization
        self.disc2 = sampler2.discretization

    def test_rv(self):
        """
        Test plotting random variable probability.
        """
        calculateR.invert_to_random_variable(self.disc1, rv='beta')
        param_labels = [r'$a$', r'$b$', r'$c$']
        for i in range(3):
            plotP.plot_1d_marginal_densities(sets=(self.disc1, self.disc2), i=i,
                                sets_label_initial=['Initial', 'Data-Generating'], sets_label=['Updated', ''],
                                title="Fitted Beta Distribution", label=param_labels[i], interactive=False)



# Copyright (C) 2014-2020 The BET Development Team

"""
This module contains unittests for :mod:`~bet.sampling.basicSampling:`
"""

import unittest
import os
import pyDOE
import numpy.testing as nptest
import numpy as np
import scipy.io as sio
import bet
import bet.sampling.basicSampling as bsam
from bet.Comm import comm
import bet.sample
from bet.sample import sample_set
from bet.sample import discretization as disc
import collections
from test.problem_setups import *

local_path = os.path.join(".")


class Test_random_sample_set1to1(unittest.TestCase):
    """
    Testing ``bet.sampling.basicSampling.random_sample_set``
    """
    def setUp(self):
        self.input_dim = 1
        self.output_dim = 1
        self.num = 100
        self.set = random_voronoi(rv='uniform', dim=self.input_dim, out_dim=self.output_dim,
                                  num_samples=self.num, level=1)

    def test_nums(self):
        """
        Test for correct dimensions and sizes.
        """
        assert self.set.get_dim() == self.input_dim
        assert self.set.get_values().shape[0] == self.num

class Test_random_sample_set3to2(Test_random_sample_set1to1):
    """
    Testing ``bet.sampling.basicSampling.random_sample_set``
    """
    def setUp(self):
        self.input_dim = 3
        self.output_dim = 2
        self.num = 100
        self.set = random_voronoi(rv=['beta', {'a': 1, 'b': 3}], dim=self.input_dim, out_dim=self.output_dim,
                                  num_samples=self.num, level=1)


class Test_random_sample_set2to1(Test_random_sample_set1to1):
    """
    Testing ``bet.sampling.basicSampling.random_sample_set``
    """
    def setUp(self):
        self.input_dim = 2
        self.output_dim = 1
        self.num = 1
        self.set = random_voronoi(rv=[['beta', {'a': 1, 'b': 3}], ['norm', {'scale': 3}]],
                                  dim=self.input_dim, out_dim=self.output_dim,
                                  num_samples=self.num, level=1)

class Test_regular_sample(unittest.TestCase):
    """
    Testing ``bet.sampling.basicSampling.regular_sample_set``
    """

    def setUp(self):
        self.input_dim = 2
        self.output_dim = 1
        self.num = 3
        self.set = regular_voronoi(dim=self.input_dim, out_dim=self.output_dim, num_samples_per_dim=self.num)

    def test_nums(self):
        """
        Test for correct dimensions and sizes.
        """
        assert self.set.get_dim() == self.input_dim
        assert self.set.get_values().shape[0] == self.num ** self.input_dim

    def test_domain(self):
        """
        Test that values are in correct domain.
        """
        assert np.all(np.greater_equal(self.set.get_values(), 0.0))
        assert np.all(np.less_equal(self.set.get_values(), 1.0))

class Test_lhs_sample(Test_random_sample_set1to1):
    """
    Testing ``bet.sampling.basicSampling.lhs_sample_set``
    """

    def setUp(self):
        self.input_dim = 2
        self.output_dim = 1
        self.num = 3
        self.set = lhs_voronoi(dim=self.input_dim, out_dim=self.output_dim, num_samples=self.num)

    def test_domain(self):
        """
        Test that values are in correct domain.
        """
        assert np.all(np.greater_equal(self.set.get_values(), 0.0))
        assert np.all(np.less_equal(self.set.get_values(), 1.0))


class Test_sampler(unittest.TestCase):
    """
    Testing ``bet.sampling.basicSampling.sampler``
    """
    def setUp(self):
        self.input_dim = 2
        self.output_dim = 2
        self.num = 100
        self.sampler = random_voronoi(rv='uniform', dim=self.input_dim, out_dim=self.output_dim,
                                  num_samples=self.num, level=3)

    def test_nums(self):
        """
        Test for correct dimensions and sizes.
        """
        assert self.sampler.discretization.get_input_sample_set().get_dim() == self.input_dim
        assert self.sampler.discretization.get_output_sample_set().get_dim() == self.output_dim
        assert self.sampler.discretization.check_nums() == self.num

    def test_copy(self):
        """
        Test copying
        """
        sampler2 = self.sampler.copy()
        assert sampler2.discretization == self.sampler.discretization
        assert sampler2.lb_model == self.sampler.lb_model

    def test_values(self):
        """
        Check for correct values.
        """
        nptest.assert_almost_equal(self.sampler.discretization.get_input_sample_set().get_values(),
                                   self.sampler.discretization.get_output_sample_set().get_values())


class Test_sampler_regular(Test_sampler):
    """
    Testing ``bet.sampling.basicSampling.sampler``
    """
    def setUp(self):
        self.input_dim = 2
        self.output_dim = 2
        self.num = 3
        self.sampler = regular_voronoi(dim=self.input_dim, out_dim=self.output_dim,
                                       num_samples_per_dim=self.num, level=3)

    def test_nums(self):
        """
        Test for correct dimensions and sizes.
        """
        assert self.sampler.discretization.get_input_sample_set().get_dim() == self.input_dim
        assert self.sampler.discretization.get_output_sample_set().get_dim() == self.output_dim
        assert self.sampler.discretization.check_nums() == self.num ** self.input_dim




class Test_sampler_lhs(Test_sampler):
    """
    Testing ``bet.sampling.basicSampling.sampler``
    """
    def setUp(self):
        self.input_dim = 2
        self.output_dim = 2
        self.num = 30
        self.sampler = lhs_voronoi(dim=self.input_dim, out_dim=self.output_dim, num_samples=self.num, level=3)





# Copyright (C) 2014-2020 The BET Development Team

"""
Consider a thin 2-dimensional (square) metal plate constructed by welding
together two rectangular metal plates of similar alloy types together.
The alloy types differ due to variations in the manufacturing of the
rectangular plates, so the thermal diffusivity is different on the left
and right sides of the resulting square plate.
We want to quantify uncertainties in these thermal diffusivities using
the heat equation to model experiments where the square plates are subject
to an external, localized, source at the center of the plate.
Assuming we have exactly two contact thermometers with which to record
exactly two temperature measurements during the experiment, the question
is the following: what are the optimal placements of these thermometers
in space-time?

See the manuscript at http://arxiv.org/abs/1601.06702 for details involving
this problem and the simulations used to generate the data in Section 5.1.
Here, we take the simulated data from the model problem saved as *.mat files
that contains parameter samples chosen in clusters around 16 random
points in the input space, and the corresponding QoIs (data) from the points
in space shown in Figure 6 of the manuscript at the 50 time steps of the
simulation (resulting in 1000 total different possible QoIs to choose from
in space-time).
We use the clusters of samples to compute gradients of the QoI using either
radial basis function, forward, or centered finite difference schemes.
These gradients are used to compute the average skewness in the possible 2D maps.
We then choose the optimal set of 2 QoIs to use in the inverse problem by
minimizing average skewness.
"""

import scipy.io as sio
import bet.sensitivity.gradients as grad
import bet.sensitivity.chooseQoIs as cqoi
import bet.Comm as comm
import bet.sample as sample
import numpy as np

# Select the type of finite difference scheme as either RBF, FFD, or CFD
fd_scheme = 'RBF'

# Import the data from the FEniCS simulation (RBF or FFD or CFD clusters)
if fd_scheme.upper() in ['RBF', 'FFD', 'CFD']:
    file_name = 'heatplate_2d_16clusters' + fd_scheme.upper() + '_1000qoi.mat'
    matfile = sio.loadmat(file_name)
else:
    print('no data files for selected finite difference scheme')
    exit()

# Select a subset of QoI to check for optimality
'''
In Figure 6 of the manuscript at http://arxiv.org/abs/1601.06702, we see
that there are 20 spatial points considered and 50 time steps for a total
of 1000 different QoI.
The QoI are indexed so that the QoI corresponding to indices

    (i-1)*20 to i*20

for i between 1 and 50 corresponds to the 20 labeled QoI from Figure 6
at time step i.

Using this information, we can check QoI either across the entire range
of all space-time locations (``indexstart = 0``, ``indexstop = 1000``), or,
we can check the QoI at a particular time (e.g., setting ``indexstart=0`` and
``indexstop = 20`` considers all the spatial QoI only at the first time step).

In general, ``indexstart`` can be any integer between 0 and 998  and
``indexstop`` must be at least 2 greater than ``indexstart`` (so between
2 and 1000 subject to the additional constraint that ``indexstop``
:math: `\geq` ``indexstart + 2`` to ensure that we check at least a single pair
of QoI.)
'''
indexstart = 0
indexstop = 20
qoiIndices = np.arange(indexstart, indexstop)

# Initialize the necessary sample objects
input_samples = sample.sample_set(2)
output_samples = sample.sample_set(1000)

# Set the input sample values from the imported file
input_samples.set_values(matfile['samples'])

# Set the data fromthe imported file
output_samples.set_values(matfile['data'])

# Create the cluster discretization
cluster_discretization = sample.discretization(input_samples, output_samples)

# Calculate the gradient vectors at each of the 16 centers for each of the
# QoI maps
if fd_scheme.upper() in ['RBF']:
    center_discretization = grad.calculate_gradients_rbf(cluster_discretization,
                                                         normalize=False)
elif fd_scheme.upper() in ['FFD']:
    center_discretization = grad.calculate_gradients_ffd(
        cluster_discretization)
else:
    center_discretization = grad.calculate_gradients_cfd(
        cluster_discretization)

input_samples_centers = center_discretization.get_input_sample_set()

# Choose a specific set of QoIs to check the average skewness of
index1 = 0
index2 = 4
(specific_skewness, _) = cqoi.calculate_avg_skewness(input_samples_centers,
                                                     qoi_set=[index1, index2])
if comm.rank == 0:
    print('The average skewness of the QoI map defined by indices ' + str(index1) +
          ' and ' + str(index2) + ' is ' + str(specific_skewness))

# Compute the skewness for each of the possible QoI maps determined by choosing
# any two QoI from the set defined by the indices selected by the
# ``indexstart`` and ``indexend`` values
skewness_indices_mat = cqoi.chooseOptQoIs(input_samples_centers, qoiIndices,
                                          num_optsets_return=10, measure=False)

qoi1 = skewness_indices_mat[0, 1]
qoi2 = skewness_indices_mat[0, 2]

if comm.rank == 0:
    print('The 10 smallest condition numbers are in the first column, the \
corresponding sets of QoIs are in the following columns.')
    print(skewness_indices_mat[:10, :])



# Copyright (C) 2014-2020 The BET Development Team

from dolfin import *
import numpy as np
import petsc4py
from petsc4py import PETSc
from slepc4py import SLEPc
from meshDS import*
# initialize petsc
petsc4py.init()
petsc4py.PETSc.Sys.popErrorHandler()


class projectKL(object):

    """Docstring for projectKL. """

    def __init__(self, mesh):
        """TODO: to be defined1. """
        # create meshDS obect
        self._mesh = mesh
        self.domain = meshDS(mesh)
        self.c_volume_array, self.c_centroid_array = self.domain.getElemVCArray()
        self.node_to_elem = self.domain.getNodesToElem()
        self.flag = False

    def getCovMat(self, cov_expr):
        """TODO: Docstring for getCovMat.

        :cov_expr: Expression (dolfin) as a function of
        :returns: covariance PETSC matrix cov_mat

        """
        # store the expression
        self.expr = cov_expr
        # create a PETSC matrix for cov_mat
        cov_mat = PETSc.Mat().create()
        cov_mat.setType('aij')
        cov_mat.setSizes(self.domain.getNodes(), self.domain.getNodes())
        cov_mat.setUp()

        # scalar valued function is evaluated in this variable
        cov_ij = np.empty((1), dtype=float)
        # the points to evalute the expression
        xycor = np.empty((4), dtype=float)

        print('---------------------------')
        print('---------------------------')
        print(' Building Covariance Matrix')
        print('---------------------------')
        print('---------------------------')
        # Loop through global nodes and build the matrix for i < j because of
        # symmetric nature.
        for node_i in range(0, self.domain.getNodes()):
            # global node node_i
            for node_j in range(node_i, self.domain.getNodes()):
                # global node node_j
                temp_cov_ij = 0
                for elem_i in self.node_to_elem[node_i]:
                    # elem_i : element attached to node_i
                    # x1 : x co-ordinate of the centroid of element elem_i
                    x1 = self.c_centroid_array[elem_i].x()
                    # y1 : x co-ordinate of the centroid of element elem_i
                    y1 = self.c_centroid_array[elem_i].y()
                    for elem_j in self.node_to_elem[node_j]:
                        # elem_j : element attached to node_j
                        # x2 : x co-ordinate for the centroid of element elem_j
                        x2 = self.c_centroid_array[elem_j].x()
                        # y2 : y co-ordinate for the centroid of element elem_j
                        y2 = self.c_centroid_array[elem_j].y()
                        xycor[0] = x1
                        xycor[1] = x2
                        xycor[2] = y1
                        xycor[3] = y2
                        # evaluate the expression
                        cov_expr.eval(cov_ij, xycor)
                        if cov_ij[0] > 0:
                            temp_cov_ij += (1.0 / 3) * (1.0 / 3) * cov_ij[0] * self.c_volume_array[elem_i] * \
                                self.c_volume_array[elem_j]
                            cov_mat.setValue(node_i, node_j, temp_cov_ij)
                            cov_mat.setValue(node_j, node_i, temp_cov_ij)
        cov_mat.assemblyBegin()
        cov_mat.assemblyEnd()
        print('---------------------------')
        print('---------------------------')
        print(' Finished Covariance Matrix')
        print('---------------------------')
        print('---------------------------')

        return cov_mat

    def _getBMat(self):
        """TODO: Docstring for getBmat. We are solving for CX = BX where C is the covariance matrix
        and B is just a mass matrix. Here we assemble B. This is a private function. DONT call this
        unless debuging.

        :returns: PETScMatrix B
        """

        # B matrix is just a mass matrix, can be easily assembled through fenics
        # however, the ordering in fenics is not the mesh ordering. so we build a temp matrix
        # then use the vertex to dof map to get the right ordering interms of
        # our mesh nodes
        V = FunctionSpace(self._mesh, "CG", 1)
        # Define basis and bilinear form
        u = TrialFunction(V)
        v = TestFunction(V)
        a = u * v * dx
        B_temp = assemble(a)

        B = PETSc.Mat().create()
        B.setType('aij')
        B.setSizes(self.domain.getNodes(), self.domain.getNodes())
        B.setUp()

        B_ij = B_temp.array()

        v_to_d_map = vertex_to_dof_map(V)

        print('---------------------------')
        print('---------------------------')
        print(' Building Mass Matrix ')
        print('---------------------------')
        print('---------------------------')
        for node_i in range(0, self.domain.getNodes()):
            for node_j in range(node_i, self.domain.getNodes()):
                B_ij_nodes = B_ij[v_to_d_map[node_i], v_to_d_map[node_j]]
                if B_ij_nodes > 0:
                    B.setValue(node_i, node_j, B_ij_nodes)
                    B.setValue(node_j, node_i, B_ij_nodes)

        B.assemblyBegin()
        B.assemblyEnd()
        print('---------------------------')
        print('---------------------------')
        print(' Finished Mass Matrix ')
        print('---------------------------')
        print('---------------------------')
        return B

    def projectCovToMesh(self, num_kl, cov_expr):
        """TODO: Docstring for projectCovToMesh. Solves CX = BX where C is the covariance matrix
        :num_kl : number of kl exapansion terms needed
        :returns: TODO

        """
        # turn the flag to true
        self.flag = True
        # get C,B matrices
        C = self.getCovMat(cov_expr)
        B = self._getBMat()
        # Solve the generalized eigenvalue problem
        eigensolver = SLEPc.EPS()
        eigensolver.create()
        eigensolver.setOperators(C, B)
        eigensolver.setDimensions(num_kl)
        eigensolver.setProblemType(SLEPc.EPS.ProblemType.GHEP)
        eigensolver.setFromOptions()
        eigensolver.solve()
        # Get the number of eigen values that converged.
        #nconv = eigensolver.get_number_converged()

        # Get N eigenpairs where N is the number of KL expansion and check if N < nconv otherwise you had
        # really bad matrix

        # create numpy array of vectors and eigenvalues
        self.eigen_funcs = np.empty((num_kl), dtype=object)
        self.eigen_vals = np.empty((num_kl), dtype=float)

        # store the eigenvalues and eigen functions
        V = FunctionSpace(self._mesh, "CG", 1)
        x_real = PETSc.Vec().create()
        x_real.setSizes(self.domain.getNodes())
        x_real.setUp()
        x_real.setValues(np.arange(0, self.domain.getNodes()),
                         np.zeros(self.domain.getNodes()))
#        for i in range(0, self.domain.getNodes()):
#            x_real.setValue(i, 0, 0.)

        for eigen_pairs in range(0, num_kl):
            lam = eigensolver.getEigenpair(eigen_pairs, x_real)
            self.eigen_funcs[eigen_pairs] = Function(V)
            # use dof_to_vertex map to map values to the function space
            self.eigen_funcs[eigen_pairs].vector()[:] = x_real.getValues(
                dof_to_vertex_map(V).astype('int32'))
            # divide by norm to make the unit norm again
            self.eigen_funcs[eigen_pairs].vector()[:] = self.eigen_funcs[eigen_pairs].vector()[:] / \
                norm(self.eigen_funcs[eigen_pairs])
            self.eigen_vals[eigen_pairs] = lam.real



#! /usr/bin/env python

# Copyright (C) 2014-2020 The BET Development Team

r"""
This example requires the following external packages not shipped
with BET:

  (1)  An installation of FEniCS that can be run using the same
  python as used for installing BET. See http://fenicsproject.org/
  for more information.

  (2) A copy of Launcher. See https://github.com/TACC/launcher for
  more information. The user needs to set certain environment
  variables inside of lbModel.py for this to run. See lbModel.py
  for more information.

This example generates samples for a KL expansion associated with
a covariance defined by ``cov`` in computeSaveKL.py on an L-shaped mesh
that defines the permeability field for a Poisson equation solved in
myModel.py.

The quantities of interest (QoI) are defined as two spatial
averages of the solution to the PDE.

The user defines the dimension of the parameter space (corresponding
to the number of KL terms) and the number of samples in this space.
"""

import numpy as np
import bet.calculateP.simpleFunP as simpleFunP
import bet.calculateP.calculateP as calculateP
import bet.postProcess.plotP as plotP
import bet.postProcess.plotDomains as plotD
import bet.sample as samp
import bet.sampling.basicSampling as bsam
from lbModel import lb_model
from myModel import my_model
from Compute_Save_KL import computeSaveKL

# Interface BET to the model.
sampler = bsam.sampler(lb_model)

# Define the number of KL terms to use to represent permeability field
num_KL_terms = 2
# Compute and save the KL expansion -- can comment out after running once
computeSaveKL(num_KL_terms)

# Initialize input parameter sample set object
input_samples = samp.sample_set(num_KL_terms)

# Set parameter domain
KL_term_min = -3.0
KL_term_max = 3.0
input_samples.set_domain(np.repeat([[KL_term_min, KL_term_max]],
                                   num_KL_terms,
                                   axis=0))

'''
Suggested changes for user:

Try with and without random sampling.

If using regular sampling, try different numbers of samples
per dimension (be careful in the dimension is not 2).
'''
# Generate samples on the parameter space
randomSampling = False
if randomSampling is True:
    input_samples = sampler.random_sample_set(
        'random', input_samples, num_samples=1E2)
else:
    input_samples = sampler.regular_sample_set(
        input_samples, num_samples_per_dim=[10, 10])

'''
A standard Monte Carlo (MC) assumption is that every Voronoi cell
has the same volume. If a regular grid of samples was used, then
the standard MC assumption is true.
'''
MC_assumption = True
# Estimate volumes of Voronoi cells associated with the parameter samples
if MC_assumption is False:
    input_samples.estimate_volume(n_mc_points=1E5)
else:
    input_samples.estimate_volume_mc()

# Create the discretization object using the input samples
my_discretization = sampler.compute_qoi_and_create_discretization(
    input_samples, savefile='FEniCS_Example.txt.gz')

'''
Suggested changes for user:

Try different reference parameters.
'''
# Define the reference parameter
#param_ref = np.zeros((1,num_KL_terms))
param_ref = np.ones((1, num_KL_terms))

# Compute the reference QoI
Q_ref = my_model(param_ref)

# Create some plots of input and output discretizations
plotD.scatter_2D(input_samples, ref_sample=param_ref[0, :],
                 filename='FEniCS_ParameterSamples.png')
if Q_ref.size == 2:
    plotD.show_data_domain_2D(my_discretization, Q_ref=Q_ref[0, :],
                              file_extension=".png")

'''
Suggested changes for user:

Try different ways of discretizing the probability measure on D defined
as a uniform probability measure on a rectangle or interval depending
on choice of QoI_num in myModel.py.
'''
randomDataDiscretization = False
if randomDataDiscretization is False:
    simpleFunP.regular_partition_uniform_distribution_rectangle_scaled(
        data_set=my_discretization, Q_ref=Q_ref[0, :], rect_scale=0.1,
        cells_per_dimension=3)
else:
    simpleFunP.uniform_partition_uniform_distribution_rectangle_scaled(
        data_set=my_discretization, Q_ref=Q_ref[0, :], rect_scale=0.1,
        M=50, num_d_emulate=1E5)

# calculate probabilities
calculateP.prob(my_discretization)

########################################
# Post-process the results
########################################
# calculate 2d marginal probs
(bins, marginals2D) = plotP.calculate_2D_marginal_probs(input_samples,
                                                        nbins=20)
# smooth 2d marginals probs (optional)
marginals2D = plotP.smooth_marginals_2D(marginals2D, bins, sigma=0.5)

# plot 2d marginals probs
plotP.plot_2D_marginal_probs(marginals2D, bins, input_samples, filename="FEniCS",
                             lam_ref=param_ref[0, :], file_extension=".png",
                             plot_surface=False)

# calculate 1d marginal probs
(bins, marginals1D) = plotP.calculate_1D_marginal_probs(input_samples,
                                                        nbins=20)
# smooth 1d marginal probs (optional)
marginals1D = plotP.smooth_marginals_1D(marginals1D, bins, sigma=0.5)
# plot 2d marginal probs
plotP.plot_1D_marginal_probs(marginals1D, bins, input_samples, filename="FEniCS",
                             lam_ref=param_ref[0, :], file_extension=".png")



# Copyright (C) 2014-2020 The BET Development Team

import bet.sample as sample
import bet.sampling.basicSampling as bsam
import numpy as np


def unit_center_set(dim=1, num_samples=100,
                    delta=1, reg=False):
    r"""
    First define a unit hyper-box sample set. Then, construct a hyper-box
    centered at ``np.array([[0.5]]*dim)`` with side-lengths ``delta``
    of this domain that has probability 1.
    (Useful for testing).

    :param int dim: dimension
    :param int num_samples: number of samples
    :param float delta: sidelength of region with positive probability
    :param bool reg: regular sampling (`num_samples` = per dimension)
    :rtype: :class:`bet.sample.sample_set`
    :returns: sample set object

    """
    s_set = sample.sample_set(dim)
    s_set.set_domain(np.array([[0, 1]] * dim))
    if reg:
        s = bsam.regular_sample_set(s_set, num_samples)
    else:
        s = bsam.random_sample_set('r', s_set, num_samples)
    dd = delta / 2.0
    if dim > 1:
        probs = 1 * (np.sum(np.logical_and(s._values <= (0.5 + dd),
                                           s._values >= (0.5 - dd)), axis=1)
                     >= dim)
    else:
        probs = 1 * (np.logical_and(s._values <= (0.5 + dd),
                                    s._values >= (0.5 - dd)))
    s.set_probabilities(probs / np.sum(probs))  # uniform probabilities
    s.set_prob_type('voronoi')
    s.estimate_volume_mc()
    s.global_to_local()
    return s


def unit_bottom_set(dim=1, num_samples=100,
                    delta=1, reg=False):
    r"""
    First define a unit hyper-box sample set. Then, construct a hyper-box
    with a corner at ``np.array([[0.0]]*dim)`` with side-lengths ``delta``
    inside of this domain that has probability 1.
    (Useful for testing).

    :param int dim: dimension
    :param int num_samples: number of samples
    :param float delta: sidelength of region with positive probability
    :param bool reg: regular sampling (`num_samples` = per dimension)
    :rtype: :class:`bet.sample.sample_set`
    :returns: sample set object

    """
    s_set = sample.sample_set(dim)
    s_set.set_domain(np.array([[0, 1]] * dim))
    if reg:
        s = bsam.regular_sample_set(s_set, num_samples)
    else:
        s = bsam.random_sample_set('r', s_set, num_samples)
    dd = delta
    if dim == 1:
        probs = 1 * (s._values <= dd)
    else:
        probs = 1 * (np.sum(s._values <= dd, axis=1) >= dim)
    s.set_probabilities(probs / np.sum(probs))  # uniform probabilities
    s.set_prob_type('voronoi')
    s.estimate_volume_mc()
    s.global_to_local()
    return s


def unit_top_set(dim=1, num_samples=100,
                 delta=1, reg=False):
    r"""
    First define a unit hyper-box sample set. Then, construct a hyper-box
    with a corner at ``np.array([[1.0]]*dim)`` with side-lengths ``delta``
    inside of this domain that has probability 1.
    (Useful for testing).

    :param int dim: dimension
    :param int num_samples: number of samples
    :param float delta: sidelength of region with positive probability
    :param bool reg: regular sampling (`num_samples` = per dimension)
    :rtype: :class:`bet.sample.sample_set`
    :returns: sample set object

    """
    s_set = sample.sample_set(dim)
    s_set.set_domain(np.array([[0, 1]] * dim))
    if reg:
        s = bsam.regular_sample_set(s_set, num_samples)
    else:
        s = bsam.random_sample_set('r', s_set, num_samples)

    dd = delta
    if dim == 1:
        probs = 1 * (s._values >= (1 - dd))
    else:
        probs = 1 * (np.sum(s._values >= (1 - dd), axis=1) >= dim)
    s.set_probabilities(probs / np.sum(probs))  # uniform probabilities
    s.set_prob_type('voronoi')
    s.estimate_volume_mc()
    s.global_to_local()
    return s



# Copyright (C) 2014-2020 The BET Development Team

"""
This module contains unittests for :mod:`~bet.util`
"""

import bet.sample
import bet.util as util
from bet.Comm import comm
import numpy.testing as nptest
import numpy as np


def get_binary_rep(i, dim):
    """
    A ``dim`` bit representation of ``i`` in binary.

    :param int i: number to represent in binary
    :param int dim: number of bits to use in the representation
    :rtype: string
    :returns: string representation of binary represenation of i

    """
    short = bin(i).partition('b')[-1]
    full = '0' * (dim - len(short)) + short
    return full


def compare_to_bin_rep(xnew):
    """
    xnew[i] == get_binar_rep(i, dim)
    """
    rep_compare = np.zeros((xnew.shape[0],), np.bool)
    for i, row in enumerate(xnew):
        row_rep = ''
        for v in row:
            row_rep += str(v)
        rep_compare[i] = (row_rep == get_binary_rep(i, len(row)))
        print(rep_compare[i])
    assert np.all(rep_compare)


def test_meshgrid_ndim():
    """
    Tests :meth:`bet.util.meshgrid_ndim` for upto 10 vectors where each vector is
    equal to ``[0, 1]``.
    """
    for i in range(10):
        x = [[0, 1] for v in range(i + 1)]
        compare_to_bin_rep(util.meshgrid_ndim(x))


def test_get_global_values():
    """
    Tests :meth:`bet.util.get_global_values`.
    """
    for provide_shape in [True, False]:
        for i in range(5):
            compare_get_global_values(i, provide_shape)


def compare_get_global_values(i, provide_shape):
    """
    Compares the results of get global values for a vector of shape ``(comm.size*2,
    i)``.

    :param int i: Dimension of the vector of length ``comm.size*2``

    """
    if comm.rank == 0:
        if i == 0:
            original_array = np.array(np.random.random((comm.size * 2, )))
        else:
            original_array = np.array(np.random.random((comm.size * 2, i)))
    else:
        original_array = None
    original_array = comm.bcast(original_array)
    my_len = original_array.shape[0] // comm.size
    my_index = np.arange(0 + comm.rank * my_len, (comm.rank + 1) * my_len)
    if i == 0:
        my_array = original_array[my_index]
    else:
        my_array = original_array[my_index, :]
    if provide_shape:
        recomposed_array = util.get_global_values(
            my_array, original_array.shape)
    else:
        recomposed_array = util.get_global_values(my_array)
    nptest.assert_array_equal(original_array, recomposed_array)


def test_fix_dimensions_vector():
    """
    Tests :meth:`bet.util.fix_dimensions_vector`
    """
    values = [1, [1], np.arange(5), np.arange(
        5), np.ones((5, 1)), np.ones((5, 5))]
    shapes = [(1,), (1,), (5,), (5,), (5,), (25,)]
    for value, shape in zip(values, shapes):
        vector = util.fix_dimensions_vector(value)
        assert vector.shape == shape


def test_fix_dimensions_vector_2darray():
    """
    Tests :meth:`bet.util.fix_dimensions_vector_2darray`
    """
    values = [1, [1], np.empty((1, 1)), np.arange(
        5), np.arange(5), np.empty((5, 1))]
    shapes = [(1, 1), (1, 1), (1, 1), (5, 1), (5, 1), (5, 1)]
    for value, shape in zip(values, shapes):
        vector = util.fix_dimensions_vector_2darray(value)
        assert vector.shape == shape


def test_fix_dimensions_domain():
    """
    Tests :meth:`bet.util.fix_dimensions_domain`
    """
    values = [np.arange(2), np.empty((2,)), np.empty((2, 1)), np.empty((1, 2)),
              np.empty((5, 2)), np.empty((2, 5))]
    shapes = [(1, 2), (1, 2), (1, 2), (1, 2), (5, 2), (5, 2)]
    for value, shape in zip(values, shapes):
        vector = util.fix_dimensions_domain(value)
        assert vector.shape == shape


def test_fix_dimensions_data_nodim():
    """
    Tests :meth`bet.util.fix_dimensions_domain` when `dim` is not specified
    """
    values = [1, [1], np.arange(2), np.empty((2,)), np.empty((2, 1)), np.empty((1, 2)),
              np.empty((5, 2)), np.empty((2, 5))]
    shapes = [(1, 1), (1, 1), (2, 1), (2, 1), (2, 1), (1, 2), (5, 2), (2, 5)]
    print(len(values), len(shapes))
    for value, shape in zip(values, shapes):
        vector = util.fix_dimensions_data(value)
        print(vector, value)
        print(vector.shape, shape)
        assert vector.shape == shape


def test_fix_dimensions_data_dim():
    """
    Tests :meth`bet.util.fix_dimensions_domain` when `dim` is specified
    """
    values = [1, [1], np.arange(2), np.empty((2,)), np.empty((2, 1)), np.empty((1, 2)),
              np.empty((5, 2)), np.empty((2, 5)), np.empty((5, 2)), np.empty((2, 5))]
    shapes = [(1, 1), (1, 1), (1, 2), (1, 2), (1, 2), (1, 2), (5, 2), (5, 2), (2, 5),
              (2, 5)]
    dims = [1, 1, 2, 2, 2, 2, 2, 2, 5, 5]
    for value, shape, dim in zip(values, shapes, dims):
        vector = util.fix_dimensions_data(value, dim)
        print(vector, value)
        print(vector.shape, shape, dim)
        assert vector.shape == shape



# Copyright (C) 2014-2020 The BET Development Team

"""

This module provides methods to sample from Lp generalized normal, uniform, and
beta distributions on the nD ball.

Adapted from natter.LpSphericallySymmetric.py
https://github.com/fabiansinz/natter

"""

import numpy as np


def Lp_generalized_normal(dim, num, p=2, scale=1.0, loc=None):
    r"""

    Generate samples from an Lp generalized normal distribution.

    :param float p: :math:`0 < p \leq \infty`, p for the lp norm where
        infinity is ``numpy.inf``
    :param int dim: Dimension of the space
    :param int num: Number of samples to generate
    :param scale: Radius to scale the samples by
    :type scale: ``float``, ``int``, or :class:`numpy.ndarray`
    :param loc: Location of the center of the samples
    :type loc: :class:`numpy.ndarray` of shape (dim,)

    """
    num = int(num)
    dim = int(dim)
    p = float(p)
    z = np.random.gamma(1. / p, scale=scale, size=(num, dim))
    z = np.abs(z)**(1. / p)
    samples = z * np.sign(np.random.randn(num, dim))
    if loc is not None:
        samples = samples + loc
    return samples


def Lp_generalized_uniform(dim, num, p=2, scale=1.0, loc=None):
    r"""

    Generate samples from an Lp generalized uniform distribution.

    :param float p: :math:`0 < p \leq \infty`, p for the lp norm where
        infinity is ``numpy.inf``
    :param int dim: Dimension of the space
    :param int num: Number of samples to generate
    :param scale: Radius to scale the samples by
    :type scale: ``float``, ``int``, or :class:`numpy.ndarray`
    :param loc: Location of the center of the samples
    :type loc: :class:`numpy.ndarray` of shape (dim,)

    """
    num = int(num)
    dim = int(dim)
    if not np.isinf(p):
        p = float(p)
        # sample from a p-generalized normal with scale 1
        samples = Lp_generalized_normal(dim, num, p)
        samples_norm = np.sum(np.abs(samples)**p, axis=1)**(1. / p)
        samples = samples / np.reshape(samples_norm, (num, 1))
        r = np.random.beta(a=dim, b=1., size=(num, 1))
        samples = samples * r * scale
    else:
        samples = (np.random.random((num, dim)) - .5) * 2.0 * scale
    if loc is not None:
        samples = samples + loc
    return samples


def Lp_generalized_beta(dim, num, p=2, d=2, scale=1.0, loc=None):
    r"""

    Generate samples from an Lp generalized beta distribution. When p=d then
    this is simly the Lp generalized uniform distribution.

    :param float p: :math:`0 < p \leq \infty`, p for the lp norm where
        infinity is ``numpy.inf``
    :param float d: shape parameter
    :param int dim: Dimension of the space
    :param int num: Number of samples to generate
    :param scale: Radius to scale the samples by
    :type scale: ``float``, ``int``, or :class:`numpy.ndarray`
    :param loc: Location of the center of the samples
    :type loc: :class:`numpy.ndarray` of shape (dim,)

    """
    num = int(num)
    dim = int(dim)
    p = float(p)
    # sample from a p-generalized normal with scale 1
    samples = Lp_generalized_normal(dim, num, p)
    samples_norm = np.sum(np.abs(samples)**p, axis=1)**(1. / p)
    samples = samples / np.reshape(samples_norm, (num, 1))
    r = np.random.beta(a=dim / p, b=d / p, size=(num, 1))**(1. / p)
    samples = samples * r * scale
    if loc is not None:
        samples = samples + loc
    return samples
